package jess;

import java.beans.*;
import java.io.Serializable;
import java.lang.reflect.*;
import java.util.*;

/**
 * Manages the list of Java objects that are subject to pattern
 * matching in a Rete engine.
 * (C) 2005 Sandia National Laboratories
 */

class DefinstanceList implements Serializable, PropertyChangeListener {

    // Keys are Jess defclass names, elements are Java class names
    private HashMap m_javaClasses = new HashMap(101);
    // Keys are objects to match, elements are the facts that represent them.
    private Map m_definstances = new IdentityHashMap(101);
    // Keys are objects to match, elements are the Jess class names
    private Map m_jessClasses = new IdentityHashMap(101);
    private transient Rete m_engine;
    public static final String JAVA_OBJECT = "$JAVA-OBJECT$ ";

    DefinstanceList(Rete engine) {
        // ###
        setEngine(engine);
    }

    void setEngine(Rete engine) {
        m_engine = engine;
    }

    void clear() {
        synchronized(m_engine.getCompiler()) {
            for (Iterator it = m_definstances.keySet().iterator(); it.hasNext();)
                removePropertyChangeListener(it.next());

            m_javaClasses.clear();
            m_definstances.clear();
            m_jessClasses.clear();
        }
    }

    void reset() throws JessException {
        synchronized (m_engine.getCompiler()) {
            for (Iterator e = m_definstances.keySet().iterator(); e.hasNext();)
                reassertShadowFact(e.next(), m_engine.getGlobalContext());
        }
    }

    Value defclass(String jessName, String clazz, String parent, boolean includeMemberVariables)
            throws JessException {

        try {
            synchronized (m_engine.getCompiler()) {
                Class c = m_engine.findClass(clazz);

                Deftemplate dt = createDeftemplate(m_engine, jessName, c, parent, includeMemberVariables);

                mapDefclassName(m_engine.resolveName(jessName), c.getName());

                // Install our synthetic deftemplate
                m_engine.addDeftemplate(dt);

                return new Value(c.getName(), RU.SYMBOL);
            }
        } catch (ClassNotFoundException cnfe) {
            throw new JessException("defclass", "Class not found:", cnfe);
        }
    }

    static Deftemplate createDeftemplate(Rete engine, String jessName, Class c, String parent, boolean includeMemberVariables)
            throws JessException {

        try {
            Deftemplate dt;
            if (parent != null) {
                dt = engine.findDeftemplate(engine.resolveName(parent));
                if (dt == null)
                    throw new JessException("defclass",
                                            "No such parent template: ",
                                            parent);
                dt = new Deftemplate(jessName, JAVA_OBJECT + c.getName(), dt, engine);
            } else
                dt = new Deftemplate(jessName, JAVA_OBJECT + c.getName(), engine);


            addBeanProperties(c, dt);
            if (includeMemberVariables)
                addPublicVariableSlots(c, dt);
            addActiveInstanceSlot(dt);
            return dt;
        } catch (IntrospectionException ie) {
            throw new JessException("defclass", "Introspection error:", ie);
        }
    }

    private static void addPublicVariableSlots(Class c, Deftemplate dt) throws JessException {
        Field[] fields = c.getFields();
        Arrays.sort(fields, new Comparator() {
            public int compare(Object o1, Object o2) {
                Field f1 = (Field) o1;
                Field f2 = (Field) o2;
                return f1.getName().compareTo(f2.getName());
            }
        });

        for (int i = 0; i < fields.length; i++) {
            final Field field = fields[i];
            if (Modifier.isStatic(field.getModifiers()))
                continue;
            String name = field.getName();
            Class rt = field.getType();
            Value defaultValue = new Value(new SerializableVD(c, field, true));
            if (rt.isArray())
                dt.addMultiSlot(name, defaultValue);
            else
                dt.addSlot(name, defaultValue, getSlotType(rt));
        }
    }

    private static void addActiveInstanceSlot(Deftemplate dt) throws JessException {
        dt.addSlot("OBJECT", Funcall.NIL, "OBJECT");
    }

    private static void addBeanProperties(Class c, Deftemplate dt) throws JessException, IntrospectionException {
        // Make all the readable 'bean properties' into slots
        PropertyDescriptor[] props =
                ReflectFunctions.getPropertyDescriptors(c);

        // Sort them first
        Arrays.sort(props, new Comparator() {
            public int compare(Object o1, Object o2) {
                PropertyDescriptor p1 = (PropertyDescriptor) o1;
                PropertyDescriptor p2 = (PropertyDescriptor) o2;
                return p1.getName().compareTo(p2.getName());
            }
        });

        for (int i = 0; i < props.length; i++) {
            Method m = props[i].getReadMethod();
            if (m == null)
                continue;
            String name = props[i].getName();
            Class rt = m.getReturnType();
            Value defaultValue = new Value(new SerializablePD(c, props[i]));
            if (rt.isArray())
                dt.addMultiSlot(name, defaultValue);
            else
                dt.addSlot(name, defaultValue, getSlotType(rt));
        }
    }

    static String getSlotType(Class type) {
        if (type == int.class || type == Integer.class ||
                type == short.class || type == Short.class ||
                type == byte.class || type == Byte.class)
            return "INTEGER";
        else if (type == long.class || type == Long.class)
            return "LONG";
        else if (type == boolean.class || type == Boolean.class)
            return "SYMBOL";
        else if (type == float.class || type == Float.class ||
                type == double.class || type == Double.class)
            return "FLOAT";
        else if (type == String.class)
            return "STRING";
        else
            return "ANY";
    }

    Value definstance(String jessTypename, Object object, boolean dynamic, Context context)
            throws JessException {
        try {
            synchronized (m_engine.getCompiler()) {
                String javaTypename = jessNameToJavaName(jessTypename, m_engine);

                if (javaTypename == null)
                    throw new JessException("definstance",
                                            "Unknown object class",
                                            jessTypename);

                // Make sure we're not already matching on this object
                if (m_definstances.get(object) != null)
                    return new FactIDValue(null);


                if (!context.getEngine().findClass(javaTypename).
                        isAssignableFrom(object.getClass()))
                    throw new JessException("definstance",
                                            "Object is not instance of",
                                            javaTypename);

                // EJFH Potential race here -- what if we are notified but object not added yet?
                if (dynamic) {
                    // Add ourselves to the object as a PropertyChangeListener

                    Class pcl = context.getEngine().findClass("java.beans.PropertyChangeListener");
                    Method apcl =
                            object.getClass().getMethod("addPropertyChangeListener",
                                                        new Class[]{pcl});
                    apcl.invoke(object, new Object[]{this});
                }

                int shadowMode = dynamic ? Fact.DYNAMIC : Fact.STATIC;
                m_jessClasses.put(object, jessTypename);
                Fact fact = createNewShadowFact(object, context, shadowMode);
                fact = m_engine.assertFact(fact, context);

                return new FactIDValue(fact);
            }
        } catch (InvocationTargetException ite) {
            throw new JessException("DefinstanceList.definstance",
                                    "Cannot add PropertyChangeListener",
                                    ite.getTargetException());

        } catch (NoSuchMethodException nsm) {
            throw new JessException("DefinstanceList.definstance",
                                    "Obj doesn't accept " +
                                            "PropertyChangeListeners",
                                    nsm);
        } catch (ClassNotFoundException cnfe) {
            throw new JessException("DefinstanceList.definstance",
                                    "Class not found", cnfe);
        } catch (IllegalAccessException iae) {
            throw new JessException("DefinstanceList.definstance",
                                    "Class or method is not accessible",
                                    iae);
        }
    }

    void undefinstanceNoRetract(Object o) throws JessException {
        synchronized (m_engine.getCompiler()) {
            removePropertyChangeListener(o);
            m_definstances.remove(o);
            m_jessClasses.remove(o);
        }
    }

    Fact undefinstance(Object o) throws JessException {
        synchronized (m_engine.getCompiler()) {
            Fact f = (Fact) m_definstances.get(o);
            undefinstanceNoRetract(o);
            if (f != null)
                f = m_engine.retractNoUndefinstance(f);
            return f;
        }
    }

    private void removePropertyChangeListener(Object o) {
        try {
            Method apcl =
                    o.getClass().getMethod("removePropertyChangeListener",
                                           new Class[]{
                                                   PropertyChangeListener.class
                                           });
            apcl.invoke(o, new Object[]{this});

        } catch (Exception e) { /* whatever */
        }
    }

    Set listDefinstances() {
        synchronized (m_engine.getCompiler()) {
            // We build a new collection in case the client does something
            // destructive with this iterator (like call undefinstance)
            return new HashSet(m_definstances.keySet());
        }
    }

    Set listDefinstances(Filter filter) {
        synchronized (m_engine.getCompiler()) {
            Set set = new HashSet();
            for (Iterator it = m_definstances.keySet().iterator(); it.hasNext();) {
                Object obj = it.next();
                if (filter.accept(obj))
                    set.add(obj);
            }
            return set;
        }
    }

    boolean containsObject(Object o) {
        return m_definstances.containsKey(o);
    }

    Iterator listDefclasses() {
        synchronized (m_engine.getCompiler()) {
            return new ArrayList(m_javaClasses.keySet()).iterator();
        }
    }

    String jessNameToJavaName(String jessName, Rete engine) {
        String clazz = (String) m_javaClasses.get(jessName);
        if (clazz == null)
            clazz = (String) m_javaClasses.get(engine.resolveName(jessName));
        if (clazz == null && jessName.indexOf("::") == -1)
            clazz = (String) m_javaClasses.get("MAIN::" + jessName);
        return clazz;
    }

    private void mapDefclassName(String jessName, String className) {
        m_javaClasses.put(jessName, className);
    }

    private Fact getShadowFactForObject(Object o) throws JessException {
        synchronized (m_engine.getCompiler()) {
            Fact fact = (Fact) m_definstances.get(o);
            if (fact == null)
                throw new JessException("DefinstanceList.getShadowFactForObject",
                                        "Object not a definstance: ", o.toString());
            return fact;
        }
    }

    // For each bean property that appears to have changed in the object,
    // modify the shadow fact.

    private Fact updateMultipleSlots(Object o, Context context)
            throws JessException {

        Fact fact = getShadowFactForObject(o);

        Rete engine = context.getEngine();

        synchronized (engine.getCompiler()) {
            m_definstances.put(o, fact);

            Deftemplate deft = fact.getDeftemplate();

            ArrayList names = new ArrayList();
            ArrayList values = new ArrayList();

            try {
                for (int i = 0; i < deft.getNSlots(); i++) {
                    String name = deft.getSlotName(i);

                    if (name.equals("OBJECT"))
                        continue;

                    Value newV = getSlotValueFromObject(deft, i, engine, o).resolveValue(context);
                    Value oldV = fact.getSlotValue(name);

                    if (oldV != null && oldV.equals(newV))
                        continue;

                    names.add(name);
                    values.add(newV);
                }
                engine.modifyRegularFact(fact, (String[]) names.toArray(new String[names.size()]),
                                         (Value[]) values.toArray(new Value[values.size()]), engine, context);

            } finally {
                engine.commitActivations();
            }
        }

        return fact;
    }

    // Modify the single slot that's changed.
    private Fact updateSingleSlot(Object o,
                                  String slotName,
                                  Object newValue,
                                  Context context)
            throws JessException {

        Fact fact = getShadowFactForObject(o);

        Rete engine = context.getEngine();
        try {
            synchronized (engine.getCompiler()) {
                Deftemplate deft = fact.getDeftemplate();
                int index = deft.getSlotIndex(slotName);
                if (index == -1)
                    throw new JessException("DeftemplateList.updateSingleSlot",
                                            "No such slot " + slotName + " in template",
                                            deft.getName());

                Value newV;
                if (newValue == null)
                    newV = getSlotValueFromObject(deft, index, engine, o);
                else
                    newV = objectPropertyToSlotValue(deft, index, engine, newValue);

                try {
                    if (!fact.getSlotValue(slotName).equals(newV))
                        engine.modifyRegularFact(fact,
                                                 new String[]{slotName},
                                                 new Value[]{newV},
                                                 engine, context);
                } finally {
                    engine.commitActivations();
                }

                return fact;
            }

        } catch (IllegalArgumentException iae) {
            throw new JessException("DefinstanceList.updateShadowFact",
                                    "Invalid argument", iae);
        }
    }

    private Value objectPropertyToSlotValue(Deftemplate deft, int index, Rete engine, Object newValue)
            throws JessException {
        SerializableD pd = (SerializableD)
                deft.getSlotDefault(index).javaObjectValue(engine.getGlobalContext());
        Class rt = pd.getPropertyType(engine);
        return ReflectFunctions.objectToValue(rt, newValue);
    }

    private void reassertShadowFact(Object o, Context context)
            throws JessException {
        Fact fact = getShadowFactForObject(o);
        setAllSlotValuesFromObject(context, fact, o);
        m_engine.assertFact(fact, context);
    }

    private Fact createNewShadowFact(Object o,
                                     Context context,
                                     int shadowMode)
            throws JessException {

        synchronized(m_engine.getCompiler()) {
            Fact fact = new Fact((String) m_jessClasses.get(o), m_engine);
            fact.setShadowMode(shadowMode);
            setAllSlotValuesFromObject(context, fact, o);
            return fact;
        }
    }

    private void setAllSlotValuesFromObject(Context context, Fact fact, Object o)
            throws JessException {
        synchronized (m_engine.getCompiler()) {
            Rete engine = context.getEngine();

            fact.setSlotValue("OBJECT", new Value(o));
            m_definstances.put(o, fact);

            Deftemplate deft = fact.getDeftemplate();

            for (int i = 0; i < deft.getNSlots(); i++) {
                String name = deft.getSlotName(i);

                if (name.equals("OBJECT"))
                    continue;

                Value newV = getSlotValueFromObject(deft, i, engine, o);

                fact.setSlotValue(name, newV);
            }
        }
    }

    private Value getSlotValueFromObject(Deftemplate deft, int slotIndex, Rete engine, Object o)
            throws JessException {

        try {
            SerializableD pd = (SerializableD)
                    deft.getSlotDefault(slotIndex).javaObjectValue(engine.getGlobalContext());

            Object newValue = pd.getPropertyValue(engine, o);
            return ReflectFunctions.objectToValue(pd.getPropertyType(engine), newValue);

        } catch (InvocationTargetException ite) {
            throw new JessException("DefinstanceList.updateMultipleSlots",
                                    "Called method threw an exception",
                                    ite.getTargetException());

        } catch (IllegalAccessException iae) {
            throw new JessException("DefinstanceList.updateMultipleSlots",
                                    "Method is not accessible",
                                    iae);
        } catch (IllegalArgumentException iae) {
            throw new JessException("DefinstanceList.updateMultipleSlots",
                                    "Invalid argument", iae);
        }

    }

    Value updateObject(Object object) throws JessException {
        Fact fact = updateMultipleSlots(object, m_engine.getGlobalContext());
        return new FactIDValue(fact);
    }

    Value updateObject(Object object, String slotName) throws JessException {
        Fact fact = updateSingleSlot(object, slotName, null, m_engine.getGlobalContext());
        return new FactIDValue(fact);
    }

    public void propertyChange(PropertyChangeEvent pce) {
        Object o = pce.getSource();

        try {
            synchronized (m_engine.getCompiler()) {
                if (m_definstances.get(o) == null)
                    return;

                String changedName = pce.getPropertyName();
                Object newValue = pce.getNewValue();
                Context context = m_engine.getGlobalContext();

                if (changedName == null)
                    updateMultipleSlots(o, context);
                else
                    updateSingleSlot(o, changedName, newValue, context);
            }

        } catch (JessException re) {
            // TODO: Explain the problem
            System.out.println("Async Error: " + re);
            re.printStackTrace();
            if (re.getCause() != null)
                re.getCause().printStackTrace();
        }
    }


}

