package jess;

import java.io.*;
import java.util.*;

/**
 * A FactList is the collection that holds and processes all the Fact
 * object for a Rete engine. You could say that this is the "working
 * memory", although that would really include all the Rete network
 * join node memories, as well.
 * <P>
 * (C) 2005 Sandia National Laboratories<br>
 */


class FactList implements Serializable {

    private Map m_table = Collections.synchronizedMap(new HashMap());
    private final ArrayList m_factsToAssert = new ArrayList();
    private final ArrayList m_factsToRetract = new ArrayList();

    private int m_time = 0;
    private final LogicalSupport m_logicalSupport = new LogicalSupport(m_factsToRetract);

    private void processToken(int tag, Fact fact, Rete engine, Context context) throws JessException {
        synchronized (engine.getCompiler()) {
            assignTime(fact);
            Token t = Rete.getFactory().newToken(fact, tag);
            engine.getCompiler().getRoot().callNodeRight(tag, t, context);
        }
    }

    void updateNodes(Rete engine, Hashtable n) throws JessException {
        try {
            JessException exception = null;
            for (Iterator e = listFacts(); e.hasNext();) {
                Fact fact = (Fact) e.next();
                Token t = Rete.getFactory().newToken(fact, RU.UPDATE);
                for (Enumeration nodes = n.elements(); nodes.hasMoreElements();)
                    synchronized (engine.getCompiler()) {
                        Node node = (Node) nodes.nextElement();
                        try {
                            node.callNodeRight(RU.UPDATE, t, engine.getGlobalContext().push());
                        } catch (JessException je) {
                            exception = je;
                        }
                    }
            }
            processPendingFacts(engine);
            if (exception != null)
                throw exception;
        } finally {
            engine.commitActivations();
        }
    }


    void clear(Rete engine) throws JessException {
        processToken(RU.CLEAR, Fact.getClearFact(), engine, engine.getGlobalContext());
        m_table.clear();
        m_time = 0;
        m_factsToAssert.clear();
        m_factsToRetract.clear();
        m_logicalSupport.clear();
        m_nextFactId = 0;
    }

    int getTime() {
        return m_time;
    }

    void assignTime(Fact fact) {
        fact.updateTime(++m_time);
    }

    int doPreAssertionProcessing(Fact f) {
        return m_table.containsKey(f) ? -1 : 0;
    }

    Fact findFactByFact(Fact f) {
        return (Fact) m_table.get(f);
    }

    Fact findFactByID(int id) {
        synchronized (m_table) {
            for (Iterator it = m_table.keySet().iterator(); it.hasNext();) {
                Fact f = (Fact) it.next();
                if (f.getFactId() == id)
                    return f;
            }
            return null;
        }
    }

    void ppFacts(String name, Writer output) throws IOException {
        for (Iterator e = listFacts(); e.hasNext();) {
            Fact f = (Fact) e.next();
            if (name == null || f.getName().equals(name)) {
                output.write(f.toString());
                output.write("\n");
            }
        }
    }

    void ppFacts(Writer output) throws IOException {
        ppFacts(null, output);
    }

    Iterator listFacts() {
        synchronized (m_table) {
            return new SortedIterator(m_table);
        }
    }

    /**
     * Successively incremented ID for asserted facts.
     */
    private Object m_idLock = new String("LOCK");
    private int m_nextFactId;

    private int consumeFactId() {
        synchronized (m_idLock) {
            return m_nextFactId++;
        }
    }

    int peekFactId() {
        synchronized(m_idLock) {
            return m_nextFactId;
        }
    }

    // Process any facts that were asserted by rule LHS processing --
    // i.e., for backwards chaining -- or should be retracted, for
    // instance, because logical support was removed.
    void processPendingFacts(Rete engine) throws JessException {
        synchronized (m_factsToAssert) {
            while (m_factsToAssert.size() > 0) {
                Fact f = (Fact) m_factsToAssert.get(0);
                m_factsToAssert.remove(0);
                _assert(f, engine, engine.getGlobalContext());
            }
        }

        synchronized (m_factsToRetract) {
            while (m_factsToRetract.size() > 0) {
                Fact f = (Fact) m_factsToRetract.get(0);
                m_factsToRetract.remove(0);
                if (f.isShadow()) {
                    Object ov = f.getSlotValue("OBJECT").javaObjectValue(null);
                    engine.undefinstanceNoRetract(ov);
                }
                _retract(f, engine);
            }
        }
    }

    Fact assertFact(Fact f, Rete engine, Context context)
            throws JessException {
        synchronized (engine.getCompiler()) {
            Fact fact = _assert(f, engine, context);
            if (fact != null)
                processPendingFacts(engine);
            return fact;
        }
    }

    private Fact _assert(Fact f, Rete engine, Context context)
            throws JessException {
        synchronized (engine.getCompiler()) {
            try {
                // find any old copy
                boolean alreadyExisted = (engine.doPreAssertionProcessing(f) != 0);

                m_logicalSupport.factAsserted(context, f, alreadyExisted);

                if (alreadyExisted)
                    return null;

                f.setFactId(consumeFactId());

                engine.broadcastEvent(JessEvent.FACT, f, context);

                m_table.put(f, f);

                // Send it to the Rete network
                processToken(RU.ADD, f, engine, engine.getGlobalContext().push());

                return f;
            } finally {
                engine.commitActivations();
            }
        }
    }

    /**
     * Retract a fact.
     *
     * @param f A Fact object. Doesn't need to be the actual object
     *          that appears on the fact-list; can just be a Fact that could
     *          compare equal to one.
     * @throws JessException If anything goes wrong.
     */

    Fact retract(Fact f, Rete engine) throws JessException {
        synchronized (engine.getCompiler()) {
            Fact ff;
            if ((ff = (Fact) m_table.get(f)) != null) {
                _retract(ff, engine);
                processPendingFacts(engine);
            }
            return ff;
        }
    }

    // f should be the actual fact
    private void _retract(Fact f, Rete engine) throws JessException {
        synchronized (engine.getCompiler()) {
            try {
                engine.broadcastEvent(JessEvent.FACT | JessEvent.REMOVED, f, null);
                m_table.remove(f);
                m_logicalSupport.removeAllLogicalSupportFor(f);
                processToken(RU.REMOVE, f, engine, engine.getGlobalContext().push());
            } finally {
                engine.commitActivations();
            }
        }
    }


    Fact modify(Fact fact, String[] slotNames, Value[] slotValues, Context context, Rete engine)
            throws JessException {

        if (slotNames.length != slotValues.length)
            throw new JessException("modify",
                                    "Wrong number of values, expected " + slotNames.length + ", got",
                                    slotValues.length);
        synchronized (engine.getCompiler()) {
            try {
                switch (fact.getShadowMode()) {

                    case Fact.NO:
                        fact = modifyRegularFact(fact, slotNames, slotValues, engine, context);
                        break;
                    case Fact.DYNAMIC:
                        fact = modifyDefinstancedObject(fact, slotNames, slotValues, engine, context);
                        break;
                    case Fact.STATIC:
                        slotValues = (Value[]) slotValues.clone();
                        for (int i = 0; i < slotValues.length; i++) {
                            slotValues[i] = slotValues[i].resolveValue(context);
                        }
                        modifyDefinstancedObject(fact, slotNames, slotValues, engine, context);
                        fact = modifyRegularFact(fact, slotNames, slotValues, engine, context);
                        break;
                    default:
                        throw new JessException("modify",
                                                "Impossible shadow mode",
                                                fact.getShadowMode());
                }
            } finally {
                engine.commitActivations();
            }
        }

        return fact;
    }

    private Fact prepareToModifyRegularFact(Fact fact, String[] slotNames, Rete engine)
            throws JessException {

        synchronized (engine.getCompiler()) {
            Context context = engine.getGlobalContext().push();
            if (fact.getDeftemplate().isSlotSpecific()) {
                context.setSlotSpecificModifiedFact(fact);
                context.setModifiedSlots(slotNames);
            }
            processToken(RU.MODIFY_REMOVE, fact, engine, context);
            fact = (Fact) m_table.remove(fact);
            m_logicalSupport.removeAllLogicalSupportFor(fact);
            return fact;
        }
    }

    private Fact finishModifyRegularFact(Fact fact, Rete engine, String[] slotNames, Context context)
            throws JessException {

        synchronized (engine.getCompiler()) {
            engine.broadcastEvent(JessEvent.FACT | JessEvent.MODIFIED, fact, context);
            m_logicalSupport.factAsserted(context, fact, false);
            if (engine.doPreAssertionProcessing(fact) == 0) {
                m_table.put(fact, fact);
                Context newContext = engine.getGlobalContext().push();
                if (fact.getDeftemplate().isSlotSpecific()) {
                    newContext.setSlotSpecificModifiedFact(fact);
                    newContext.setModifiedSlots(slotNames);
                }
                processToken(RU.MODIFY_ADD, fact, engine, newContext);
                processPendingFacts(engine);
                return fact;
            } else
                return Fact.getNullFact();
        }
    }

    Fact modifyRegularFact(Fact fact, String[] slotNames, Value[] slotValues, Rete engine, Context context)
            throws JessException {

        synchronized (engine.getCompiler()) {
            if (m_table.get(fact) == null)
                throw new JessException("modify", "Fact object not in working memory", fact.toStringWithParens());

            List nameList = new ArrayList();
            List valueList = new ArrayList();
            for (int i=0; i < slotValues.length; i++) {
                Value v = slotValues[i].resolveValue(context);
                if (!fact.getSlotValue(slotNames[i]).equals(v)) {
                    nameList.add(slotNames[i]);
                    valueList.add(v);
                }
            }
            if (nameList.size() == 0)
                return fact;
            slotNames = (String[]) nameList.toArray(new String[nameList.size()]);
            fact = prepareToModifyRegularFact(fact, slotNames, engine);
            try {
                for (int i = 0; i < nameList.size(); i++) {
                    fact.setSlotValue((String) nameList.get(i), (Value) valueList.get(i));
                }
            } finally {
                fact = finishModifyRegularFact(fact, engine, slotNames, context);
            }
        }
        return fact;
    }

    private Fact modifyDefinstancedObject(Fact fact, String[] slotNames, Value[] slotValues, Rete engine,
                                          Context context)
            throws JessException {

        try {
            Deftemplate deft = fact.getDeftemplate();
            Object ov = fact.getSlotValue("OBJECT").javaObjectValue(context);
            for (int i = 0; i < slotNames.length; i++) {
                int index = deft.getSlotIndex(slotNames[i]);
                Value defaultValue = deft.getSlotDefault(index);
                SerializableD pd = (SerializableD) defaultValue.javaObjectValue(context);
                Class type = pd.getPropertyType(engine);
                pd.setPropertyValue(engine, ov, ReflectFunctions.valueToObject(type, slotValues[i], context));
            }

        } catch (Exception e) {
            throw new JessException("modify", "Error setting slot value", e);
        }
        return fact;
    }

    public List getSupportedFacts(Fact supporter) {
        return m_logicalSupport.getSupportedFacts(supporter);
    }

    public List getSupportingTokens(Fact fact) {
        return m_logicalSupport.getSupportingTokens(fact);
    }

    public void removeLogicalSupportFrom(Token token, Fact fact) {
        m_logicalSupport.removeLogicalSupportFrom(token, fact);
    }

    public void setPendingFact(Fact fact, boolean assrt) {
        if (assrt)
            m_factsToAssert.add(fact);

        else {
            fact = (Fact) m_table.get(fact);
            if (fact != null)
                m_factsToRetract.add(fact);
        }
    }
}

class FactComparator implements Comparator, Serializable {
    public int compare(Object o1, Object o2) {
        return ((Fact) o1).getFactId() - ((Fact) o2).getFactId();
    }
}

class SortedIterator implements Iterator, Serializable {
    private Object[] m_arr;
    private int m_index;

    SortedIterator(Map map) {
        m_arr = map.keySet().toArray();
        Arrays.sort(m_arr, new FactComparator());
    }

    public boolean hasNext() {
        return m_index < m_arr.length;
    }

    public Object next() {
        return m_arr[m_index++];
    }

    public void remove() throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }
}



