package jess;

import jess.awt.TextReader;
import jess.factory.Factory;
import jess.factory.FactoryImpl;
import jess.server.LineNumberRecord;

import java.applet.Applet;
import java.io.*;
import java.net.URL;
import java.util.*;

/**
 <p>The central class in the Jess library. Executes the built Rete network,
 and coordinates many
 other activities. Rete is basically a facade for all the other classes
 in the Jess library.</p>


 <p>The <tt>jess.Rete</tt> class is the rule engine itself. Each
 <tt>jess.Rete</tt> object has its own working memory, agenda, rules,
 etc. To embed Jess in a Java application, you'll simply need to create
 one or more <tt>jess.Rete</tt> objects and manipulate them
 appropriately.</p>

 (C) 2005 Sandia National Laboratories<br>
 */

public class Rete implements Serializable, JessListener {

    private Context m_globalContext = new Context(this);
    private transient Routers m_routers = new Routers();
    private transient TextReader m_tis = new TextReader(true);
    private transient Jesp m_jesp = initInternalParser();
    private transient JessEventSupport m_jes = new JessEventSupport(this);
    private boolean m_resetGlobals = true;
    private java.util.List m_deffacts = Collections.synchronizedList(new ArrayList());
    private java.util.List m_defglobals = Collections.synchronizedList(new ArrayList());
    private Map m_functions = Collections.synchronizedMap(new HashMap(101));
    private FactList m_factList = new FactList();
    private DefinstanceList m_definstanceList = new DefinstanceList(this);
    private Map m_rules = Collections.synchronizedMap(new TreeMap());
    private ReteCompiler m_compiler = new ReteCompiler();
    private Map m_storage = Collections.synchronizedMap(new HashMap());
    private static Factory s_factory = new FactoryImpl();
    private Agenda m_agenda = new Agenda(this);
    private boolean[] m_watchInfo = new boolean[5];
    private static final String LIBRARY_NAME = "scriptlib.clp";
    private ClassSource m_classSource;
    private static Map s_lineNumberTable;
    private boolean m_passiveMode;
    private String m_watchRouter = "WSTDOUT";
    private HashSet m_features = new HashSet();

    /**
     * Constant passed to setEvalSalience() to indicate that dynamic
     * salience expressions should be evaluated only when a rule is
     * installed (the default.)
     */

    public static final int INSTALL = 0;

    /**
     * Constant passed to setEvalSalience() to indicate that dynamic
     * salience expressions should be evaluated whenever a rule is
     * activated.
     */

    public static final int ACTIVATE = 1;

    /**
     * Constant passed to setEvalSalience() to indicate that dynamic
     * salience expressions should be evaluated before every rule is
     * fired. This has a large impact on performance.
     */

    public static final int EVERY_TIME = 2;

    // **********************************************************************
    // Constructors and pseudoconstructors
    // **********************************************************************

    /**
     * Construct a single rule engine instance. It will use the
     * context ClassLoader to load classes named in Jess code.
     */
    public Rete() {
        this(null);
    }

    /**
     * Construct a single rule engine instance for use in an
     * applet. It will use the applet's ClassLoader to load classes
     * named in Jess code.
     *
     * @param applet If this Rete object is being created inside
     *               an applet, pass it as an argument.
     */
    public Rete(Applet applet) {
        this((Object) applet);
    }


    /*
     * Construct a Rete object and supply an example of an application
     * class.  Jess will try to find things (like files passed as
     * arguments to "batch") "near" where this class was loaded; i.e.,
     * it will find them using the classloader used to load the argument.
     * @param appObject any object loaded by the application-level class loader
     */

    public Rete(Object appObject) {
        // ###

        try {
            m_classSource = new ClassSource(appObject, this);
            Deftemplate.addStandardTemplates(this);
        } catch (JessException je) {
            throw new RuntimeException(je);
        }

        setEventMask(0);

        loadScriptlib();
    }

    // **********************************************************************
    // I/O Router functions
    // **********************************************************************

    /**
     * Add an "input router" to this object. An input router is
     * basically just a Reader that Jess can read from. It's stored in
     * a map and looked up by name. The router "t" corresponds to
     * standard input, but you can define your own routers, or
     * reassign existing ones, using this method.
     *
     * @param s           the router name
     * @param is          a Reader where the router's data comes from
     * @param consoleLike see the Jess manual
     */
    public void addInputRouter(String s, Reader is, boolean consoleLike) {
        m_routers.addInputRouter(s, is, consoleLike);
    }

    /**
     * Remove the named router from the engine. The Reader is not
     * closed -- if it needs to be closed, you must do that yourself.
     *
     * @param s the name of the router to remove
     */
    public void removeInputRouter(String s) {
        m_routers.removeInputRouter(s);
    }

    /**
     * Return the Reader registered under a given router name.
     *
     * @param s the router name
     * @return the router, or null if none
     */
    public Reader getInputRouter(String s) {
        return m_routers.getInputRouter(s);
    }

    Tokenizer getInputWrapper(Reader is) {
        return m_routers.getInputWrapper(is);
    }

    /**
     * Add an "output router" to this object. An input router is
     * basically just a Writer that Jess can send data to. It's stored
     * in a map and looked up by name. The router "t" corresponds to
     * standard output, but you can define your own routers, or
     * reassign existing ones, using this method.
     *
     * @param s  the router name
     * @param os where the data should go
     */
    public void addOutputRouter(String s, Writer os) {
        m_routers.addOutputRouter(s, os);
    }

    /**
     * Remove the named router from the engine. The Writer is not
     * closed -- if it needs to be closed, you must do that yourself.
     *
     * @param s the name of the router
     */
    public void removeOutputRouter(String s) {
        m_routers.removeOutputRouter(s);
    }

    /**
     * See the Jess manual for a discussion of this method.
     *
     * @param s the router name
     * @return the console-like property for that router
     */
    public boolean getInputMode(String s) {
        return m_routers.getInputMode(s);
    }

    /**
     * Return the Writer registered under a given router name.
     *
     * @param s The router name
     * @return The router, or null if none
     */
    public Writer getOutputRouter(String s) {
        return m_routers.getOutputRouter(s);
    }

    /**
     * Return the WSTDERR router, where Jess sends error messages.
     *
     * @return The WSTDERR router
     */
    public PrintWriter getErrStream() {
        return m_routers.getErrStream();
    }

    /**
     * Return the WSTDERR router, where Jess sends much of its
     * standard output. The "t" router is distinct from WSTDOUT, but
     * they initially both go to System.out.
     *
     * @return The WSTDOUT router
     */
    public PrintWriter getOutStream() {
        return m_routers.getOutStream();
    }

    // **********************************************************************
    // Working memory stuff
    // **********************************************************************

    /**
     * This method is called as a Fact is about to be asserted. It is
     * meant to be overridden by extensions that need to do additional
     * processig when facts are added to working memory.
     *
     * @param f a Fact about to be asserted
     * @return non-zero if the assertion should not proceed
     */
    protected int doPreAssertionProcessing(Fact f) throws JessException {
        return m_factList.doPreAssertionProcessing(f);
    }

    /**
     * Reinitializes this rule engine. Working memory is cleared, all
     * rules are deleted, deffunctions are removed, all templates are
     * forgotten, the watch state is reset, and a JessEvent of type
     * JessEvent.CLEAR is sent, and then the event mask is set to 0.
     *
     * @throws JessException if anything goes wrong
     */
    public synchronized void clear() throws JessException {

        clearStorage();
        m_globalContext.clear();

        m_factList.clear(this);
        m_rules.clear();
        m_agenda.clear(this);
        m_definstanceList.clear();
        m_features.clear();

        keepJavaUserFunctionsOnly();
        Deftemplate.addStandardTemplates(this);

        m_compiler = new ReteCompiler();

        m_deffacts.clear();
        m_defglobals.clear();
        m_classSource.clear();

        broadcastEvent(JessEvent.CLEAR, this, null);
        setEventMask(0);

        unwatchAll();

        loadScriptlib();

        System.gc();
    }

    private void keepJavaUserFunctionsOnly() {
        ArrayList al = new ArrayList();
        for (Iterator it = m_functions.keySet().iterator(); it.hasNext();) {
            Userfunction uf = findUserfunction((String) it.next());
            if (!(uf instanceof Deffunction))
                al.add(uf);
        }

        m_functions = Collections.synchronizedMap(new HashMap());
        for (Iterator it = al.iterator(); it.hasNext();)
            addUserfunction((Userfunction) it.next());
    }

    /**
     * TODO Should this be public?
     */

    public void setPendingFact(Fact fact, boolean assrt) {
        m_factList.setPendingFact(fact, assrt);
    }

    void removeAllFacts() throws JessException {
        m_factList.clear(this);
    }

    /**
     * Reset the rule engine. Clears working memory and the agenda,
     * but rules remain defined.  Clears all non-globals from the
     * global scope. Asserts (initial-fact), reasserts all deffacts
     * and definstances, then broadcasts a JessEvent of type JessEvent.RESET.
     *
     * @throws JessException if anything goes wrong
     */

    public void reset() throws JessException {

        synchronized (m_compiler) {
            removeAllFacts();
            m_globalContext.removeNonGlobals();
            m_agenda.reset(this);
            assertFact(Fact.getInitialFact(), getGlobalContext());
            resetDefglobals();
            resetDeffacts();
        }
        m_definstanceList.reset();
        broadcastEvent(JessEvent.RESET, this, null);
        // EJFH TODO Make this switchable!
        // System.gc();
    }

    private void resetDeffacts() throws JessException {
        for (Iterator it = m_deffacts.iterator(); it.hasNext();)
            ((Deffacts) it.next()).reset(this);
    }

    private void resetDefglobals() throws JessException {
        if (getResetGlobals()) {
            for (Iterator e = m_defglobals.iterator(); e.hasNext();)
                ((Defglobal) e.next()).reset(this);
        }
    }


    /**
     * Assert a fact, as a String. For example, you can say<p/>
     * <tt>Rete engine = ...<br/>
     * Context context = ...<br/>
     * engine.assertString("(person (name Fred))", context);</tt>
     *
     * @param s a String representing a fact
     * @param c an execution context for resolving variables
     * @return the fact that was asserted
     * @throws JessException if something goes wrong
     */

    public Fact assertString(String s, Context c) throws JessException {
        synchronized (m_tis) {
            m_tis.clear();
            m_jesp.clear();
            m_tis.appendText("(assert " + s + ")");
            Value v = m_jesp.parseAndExecuteFuncall(null, c, this, m_jesp.getTokenStream(), false);
            if (v.type() == RU.FACT)
                return v.factValue(c);
            else
                return null;
        }
    }

    /**
     * Assert a fact, as a String, using the global execution
     * context. For example, you can say<p/>
     * <tt>Rete engine = ...<br/>
     * engine.assertString("(person (name Fred))");</tt>
     *
     * @param s a String representing a fact
     * @return the fact that was asserted
     * @throws JessException if something goes wrong
     */

    public Fact assertString(String s) throws JessException {
        return assertString(s, getGlobalContext());
    }

    /**
     * Assert a fact. Adds the given Fact to working memory. This fact
     * becomes the property of Jess after calling assertFact() --
     * don't change any of its fields until the fact is retracted!
     *
     * @param f a Fact object.
     * @return the fact
     * @throws JessException if anything goes wrong
     */

    public Fact assertFact(Fact f) throws JessException {
        return assertFact(f, getGlobalContext());
    }

    /**
     * Assert a fact, using the given execution context. Adds the
     * given Fact to working memory. This fact becomes the property of
     * Jess after calling assertFact() -- don't change any of its
     * fields until the fact is retracted!
     *
     * @param f a Fact object.
     * @param c an execution context
     * @return the fact
     * @throws JessException if anything goes wrong
     */

    public Fact assertFact(Fact f, Context c) throws JessException {
        return m_factList.assertFact(f, this, c);
    }

    /**
     * Retract a fact represented as a string. Parses the String to
     * create a Fact object, then tries to remove that fact from
     * working memory.
     *
     * @param s a String form of a Fact
     * @throws JessException
     */
    public Fact retractString(String s) throws JessException {
        try {
            synchronized (m_tis) {
                m_tis.clear();
                m_jesp.clear();
                m_tis.appendText(s);
                Fact f = m_jesp.parseFact(this, m_jesp.getTokenStream());
                return retract(f);
            }
        } catch (Exception t) {
            throw new JessException("Rete.retractString", s, t);
        }
    }

    /**
     * Retract a fact. Removes the Fact from working memory. Doesn't
     * need to be the actual object that appears on the fact-list; can
     * just be a Fact that could compare equal to one.
     *
     * @param f a Fact object
     * @throws JessException if anything goes wrong
     */

    public Fact retract(Fact f) throws JessException {
        if (f.isShadow()) {
            Object ov = f.getSlotValue("OBJECT").
                    javaObjectValue(null);
            return undefinstance(ov);
        } else {
            return m_factList.retract(f, this);
        }
    }

    Fact retractNoUndefinstance(Fact f) throws JessException {
        return m_factList.retract(f, this);
    }

    /**
     * Modify one slot in a fact. This function works for both
     * plain and shadow facts. For shadow facts, the corresponding
     * Java object is also modified. Uses the global context to
     * resolve the value.
     *
     * @param fact  a fact that's currently in working memory
     * @param slotName  the name of a slot in the fact
     * @param slotValue a new value for the named slot
     * @return the fact argument
     * @throws JessException if the slot name is bad or any other error occurs
     */
    public Fact modify(Fact fact, String slotName, Value slotValue) throws JessException {
        return modify(fact, new String[]{slotName}, new Value[] {slotValue});
    }


    /**
     * Modify one slot in a fact. This function works for both
     * plain and shadow facts. For shadow facts, the corresponding
     * Java object is also modified. Uses the supplied context to
     * resolve the value.
     *
     * @param fact  a fact that's currently in working memory
     * @param slotName  the name of a slot in the fact
     * @param slotValue a new value for the named slot
     * @param context an execution context
     * @return the fact argument
     * @throws JessException if the slot name is bad or any other error occurs
     */
    public Fact modify(Fact fact, String slotName, Value slotValue, Context context) throws JessException {
        return modify(fact, new String[]{slotName}, new Value[] {slotValue}, context);
    }

    /**
     * Modify any number of slots in a fact. This function works for both
     * plain and shadow facts. For shadow facts, the corresponding
     * Java object is also modified. Uses the global context to
     * resolve the values.
     *
     * @param fact  a fact that's currently in working memory
     * @param slotNames  the names of some slots in the fact
     * @param slotValues new values for the named slots
     * @return the fact argument
     * @throws JessException if any slot name is bad or any other error occurs
     */
    public Fact modify(Fact fact, String[] slotNames, Value[] slotValues) throws JessException {
        return modify(fact, slotNames, slotValues, m_globalContext);
    }

    /**
     * Modify any number of slots in a fact. This function works for both
     * plain and shadow facts. For shadow facts, the corresponding
     * Java object is also modified. Uses the given context to
     * resolve the values.
     *
     * @param fact    a fact that's currently in working memory.
     * @param slotNames    the names of some slots in the fact
     * @param slotValues   new values for the named slots.
     * @param context the execution context
     * @return the fact argument
     * @throws JessException if any slot name is bad or any other error occursy
     */
    public Fact modify(Fact fact, String[] slotNames, Value[] slotValues, Context context) throws JessException {
        return m_factList.modify(fact, slotNames, slotValues, context, this);
    }

    /**
     * For internal use only. Ignores the shadow-fact nature of a fact.
     */
    Fact modifyRegularFact(Fact fact, String[] slotNames, Value[] slotValues, Rete engine,
                           Context context) throws JessException {
        return m_factList.modifyRegularFact(fact, slotNames, slotValues, engine, context);
    }


    /**
     * Throws if called for shadow facts
     */

    Fact duplicate(ValueVector vv, Context context) throws JessException {
        return Duplicate.duplicate(vv, context);
    }

    /**
     * Return a Fact object given its numeric fact-id. This method is
     * very slow; don't use it unless you have to.  Consider the
     * returned Fact to be READ-ONLY!
     *
     * @param id the fact-id
     * @return the fact, or null if none
     * @throws JessException if something goes wrong
     */

    public Fact findFactByID(int id) throws JessException {
        return m_factList.findFactByID(id);
    }

    /**
     * Find a Fact object in working memory, given a Fact object that
     * is identical to it. This find is fast, and can be used to find
     * out quickly if a given fact is on the fact-list and if so,
     * obtain a reference to it. The argument doesn't have to be a
     * fact in working memory -- only a Fact object identical to one
     * that is.
     *
     * @param f a fact
     * @return a fact from working memory, or null if none
     * @throws JessException if something goes wrong
     */

    public Fact findFactByFact(Fact f) throws JessException {
        return m_factList.findFactByFact(f);
    }


    /**
     * Return a marker for the current state of working memory. The marker can later be used to
     * retract all facts asserted since the marker was placed using resetToMark().
     *
     * @see Rete#resetToMark(WorkingMemoryMarker)
     * @return a marker
     */


    public WorkingMemoryMarker mark() {
        return new IntMarkerImpl(m_factList.peekFactId());
    }

    /**
     * Restore working memory by retracting all facts asserted after the given marker was placed.
     * @see Rete#mark()
     * @param marker the marker
     * @throws JessException if anything goes wrong
     */
    public void resetToMark(WorkingMemoryMarker marker) throws JessException {
        marker.restore(this);
    }

    /**
     * Return a subset of Java objects (definstances) in working memory. The Iterator will
     * include only those objects for which the {@link Filter#accept} method returns true.
     *
     * @param filter a filter to apply to working memory
     * @return an Iterator over the selected objects
     */
    public Iterator getObjects(Filter filter) {
        return m_definstanceList.listDefinstances(filter).iterator();
    }

    // **********************************************************************
    // Storing, finding, listing
    // **********************************************************************

    /**
     * Pretty-print all the facts with the given head to the Writer.
     *
     * @param head   the name or "head" of the facts of interest
     * @param output the Writer to send the data to
     * @throws IOException if anything goes wrong
     */

    public void ppFacts(String head, Writer output) throws IOException {
        m_factList.ppFacts(resolveName(head), output);
    }

    /**
     * Pretty-print all the facts in working memory to the Writer.
     *
     * @param output the Writer to send the data to
     * @throws IOException if anything goes wrong
     */

    public void ppFacts(Writer output) throws IOException {
        m_factList.ppFacts(output);
    }

    /**
     * Return an Iterator over all the deffacts in this engine.
     *
     * @return the iterator
     */
    public Iterator listDeffacts() {
        synchronized (m_deffacts) {
            return new ArrayList(m_deffacts).iterator();
        }
    }

    /**
     * Return the named deffacts object.
     *
     * @param name the name of a deffacts construct
     * @return a Deffacts, or null if none
     */
    public Deffacts findDeffacts(String name) {
        name = resolveName(name);
        for (Iterator it = listDeffacts(); it.hasNext();) {
            Deffacts df = (Deffacts) it.next();
            if (df.getName().equals(name))
                return df;
        }
        return null;
    }

    /**
     * Return an Iterator over all the deftemplates in this engine,
     * both explicit and implied.
     *
     * @return the iterator
     */
    public Iterator listDeftemplates() {
        ArrayList al = new ArrayList();
        for (Iterator modules = m_agenda.listModules(); modules.hasNext();) {
            try {
                Defmodule module =
                        m_agenda.getModule((String) modules.next());
                for (Iterator dts = module.listDeftemplates();
                     dts.hasNext();)
                    al.add(dts.next());
            } catch (JessException je) {
                continue;
            }
        }
        return al.iterator();
    }

    /**
     * Return an Iterator over all the defrules in this engine.
     *
     * @return the iterator
     */
    public Iterator listDefrules() {
        synchronized (m_rules) {
            return new ArrayList(m_rules.values()).iterator();
        }
    }

    /**
     * Return an Iterator over all the facts currently in working memory.
     * The facts are returned in fact-id order.
     *
     * @return the iterator
     */
    public Iterator listFacts() {
        return m_factList.listFacts();
    }


    /**
     * Return an Iterator over all the Java objects currently
     * represented by shadow facts in working memory.
     *
     * @return the iterator
     */
    public Iterator listDefinstances() {
        return m_definstanceList.listDefinstances().iterator();
    }

    /**
     * Indicates whether a given object is being held in working memory. Will return true only for
     * the same physical object, and not for distinct but equal() objects.
     *
     * @param o the object
     * @return true if the object is in working memory
     */
    public boolean containsObject(Object o) {
        return m_definstanceList.containsObject(o);
    }

    /**
     * Return an Iterator over all the names of all defclasses. You
     * can use each name to look up the corresponding template using
     * findDeftemplate, or the corresponding Java class using
     * javaClassForDefclass.
     *
     * @return the iterator
     * @see #findDeftemplate
     * @see #javaClassForDefclass
     */
    public Iterator listDefclasses() {
        return m_definstanceList.listDefclasses();
    }


    /**
     * Return an Iterator over all the defglobals in this engine.
     *
     * @return the iterator
     */
    public Iterator listDefglobals() {
        synchronized (m_defglobals) {
            return new ArrayList(m_defglobals).iterator();
        }
    }

    /**
     * Return an Iterator over all user-defined functions and
     * deffunctions in this engine.
     *
     * @return the iterator
     */
    public Iterator listFunctions() {

        // Strip advice and FunctionHolders here.
        ArrayList v = new ArrayList();
        synchronized (m_functions) {
            for (Iterator e = m_functions.keySet().iterator(); e.hasNext();)
                v.add(findUserfunction((String) e.next()));
        }

        return v.iterator();
    }

    int countFunctions() {
        return m_functions.size();
    }

    /**
     * Find a defrule or defquery object with a certain name.
     *
     * @param name the name
     * @return the found rule or query, or null.
     */
    public final HasLHS findDefrule(String name) {
        return (HasLHS) m_rules.get(resolveName(name));
    }

    /**
     * Return the Java Class corresponding to a given Defclass name,
     * or null if the name was not found.
     *
     * @param name the name of a defclass
     * @return the Class object used to define that defclass
     */

    public Class javaClassForDefclass(String name) {
        try {
            String clazz = m_definstanceList.jessNameToJavaName(name, this);
            if (clazz == null)
                return null;
            else
                return findClass(clazz);
        } catch (ClassNotFoundException cnfe) {
            return null;
        }
    }

    /**
     * Find a deftemplate object with a certain name.
     *
     * @param name the name
     * @return the deftemplate, or null
     */
    public Deftemplate findDeftemplate(String name) throws JessException {
        if (Deftemplate.isSpecialName(name))
            return Deftemplate.getSpecialTemplate(name);

        String fullName = resolveName(name);
        String moduleName = fullName.substring(0, fullName.indexOf("::"));
        Deftemplate deft = m_agenda.getModule(moduleName).getDeftemplate(name);

        if (deft == null && !fullName.equals(name)) {
            // It may actually be defined in MAIN
            deft = m_agenda.getModule(Defmodule.MAIN).getDeftemplate(name);
        }

        return deft;
    }

    /**
     * Find or create the deftemplate by the given name. If there is
     * no such deftemplate, an ordered template is created.
     *
     * @param name the "head" of the template
     * @return a Deftemplate object
     * @throws JessException if the template name is invalid
     */
    public Deftemplate createDeftemplate(String name)
            throws JessException {

        Deftemplate deft = findDeftemplate(name);

        if (deft == null) {
            // this is OK. Create an implied deftemplate
            deft = addDeftemplate(new Deftemplate(name, "(Implied)", this));
            deft.addMultiSlot(RU.DEFAULT_SLOT_NAME, Funcall.NILLIST);

        }

        return deft;
    }

    /**
     * Creates a new deftemplate in this rule engine.  Ensures that
     * every deftemplate has a unique name by rejecting duplicates.
     *
     * @param dt a new Deftemplate
     * @return the argument
     * @throws JessException if the deftemplate is already defined
     */
    public Deftemplate addDeftemplate(Deftemplate dt) throws JessException {
        Defmodule module = m_agenda.getModule(dt.getModule());
        // setCurrentModule(module.getName());
        return module.addDeftemplate(dt, this);
    }

    /**
     * Add a deffacts construct to this rule engine.
     *
     * @param df a new Deffacts object
     * @return the argument
     * @throws JessException If an error occurs during event broadcasting
     */
    public Deffacts addDeffacts(Deffacts df) throws JessException {
        broadcastEvent(JessEvent.DEFFACTS, df, null);
        m_deffacts.add(df);
        return df;
    }

    /**
     * Creates a new Defglobal in this rule engine. This defglobal
     * will be reset immediately, regardless of the setting of
     * resetGlobals.
     *
     * @param dg a new Defglobal object
     * @return the argument
     * @throws JessException if an error occurs
     */
    public Defglobal addDefglobal(Defglobal dg) throws JessException {
        broadcastEvent(JessEvent.DEFGLOBAL, dg, null);

        dg.reset(this);
        m_defglobals.add(dg);
        return dg;
    }

    /**
     * Adds a list of Defglobals to this object, as if by addDefglobal().
     *
     * @param dg a List of Defglobals
     * @return the argument
     * @throws JessException if an error occurs
     */
    public List addDefglobals(List dg) throws JessException {
        for (Iterator it = dg.iterator(); it.hasNext();)
            addDefglobal((Defglobal) it.next());
        return dg;
    }

    /**
     * Look up a defglobal by name. The name should include the
     * leading and trailing asterisk required for defglobals.
     *
     * @param name the name of the defglobal
     * @return the Defglobal, if found, or null
     */
    public Defglobal findDefglobal(String name) {
        for (Iterator e = listDefglobals(); e.hasNext();) {
            Defglobal dg = (Defglobal) e.next();
            if (dg.getName().equals(name))
                return dg;
        }
        return null;
    }

    /**
     * Removes all user-defined functions from this engine.
     * Only the built-in functions will remain.
     */

    public void removeUserDefinedFunctions() {
        m_functions.clear();
    }

    /**
     * Creates a new function in this rule engine. Any existing
     * function by the same name will be replaced. Built-in functions
     * can be redefined this way.
     *
     * @param uf a new Userfunction
     * @return the parameter, or null if call rejected by event handler
     */
    public Userfunction addUserfunction(Userfunction uf) {
        try {
            broadcastEvent(JessEvent.USERFUNCTION, uf, null);
        } catch (JessException je) {
            return null;
        }

        FunctionHolder fh;
        if ((fh = (FunctionHolder) m_functions.get(uf.getName())) != null)
            fh.setFunction(uf);
        else
            fh = new FunctionHolder(uf);
        m_functions.put(uf.getName(), fh);
        return uf;
    }

    /**
     * Remove any currently defined function by the given
     * name. Built-in functions can be removed this way.
     *
     * @param name the name of the function to remove
     */
    public void removeUserfunction(String name) {
        Userfunction removed = findUserfunction(name);
        if (removed != null) {
            try {
                broadcastEvent(JessEvent.USERFUNCTION | JessEvent.REMOVED, removed, null);
            } catch (JessException je) {
                return;
            }
            m_functions.remove(removed);
        }
    }


    /**
     * Add a Userpackage to this engine. A Userpackage is a collection
     * of Userfunctions. A package generally calls addUserfunction
     * lots of times.
     *
     * @param up the package object
     * @return the  package object, or null if  call rejected by event
     *         handler
     */
    public Userpackage addUserpackage(Userpackage up) {
        try {
            broadcastEvent(JessEvent.USERPACKAGE, up, null);
        } catch (JessException je) {
            return null;
        }

        up.add(this);
        return up;
    }

    /**
     * Find a userfunction, if there is one by the given name.
     *
     * @param name the name of the function
     * @return the Userfunction object, if there is one, or null.
     */
    public final Userfunction findUserfunction(String name) {
        FunctionHolder fh = (FunctionHolder) m_functions.get(name);
        if (fh != null) {
            return fh.getFunction();
        } else
            return Funcall.getIntrinsic(name);
    }

    /**
     * Find a userfunction, if there is one.
     *
     * @param name The name of the function
     * @return The Userfunction object, if there is one.
     */
    final FunctionHolder findFunctionHolder(String name) {
        FunctionHolder fh = (FunctionHolder) m_functions.get(name);
        if (fh == null) {
            Userfunction uf = Funcall.getIntrinsic(name);
            if (uf != null)
                addUserfunction(uf);
            fh = (FunctionHolder) m_functions.get(name);
        }
        return fh;
    }

    /**
     * Add a rule or query to this engine. The rule is compiled and
     * added to the Rete network. Any existing rule or query by the
     * same name is removed first.
     *
     * @param dr a Defrule or Defquery
     * @return the added object
     * @throws JessException if anything goes wrong
     */
    public final HasLHS addDefrule(HasLHS dr) throws JessException {
        if (m_passiveMode)
            return dr;
        synchronized (m_compiler) {
            JessException exception = null;
            unDefrule(dr.getName());
            HasLHS hlhs = dr;
            try {
                while (hlhs != null) {
                    m_compiler.addRule(hlhs, this);
                    hlhs = hlhs.getNext();
                }

            } catch (RuleCompilerException rce) {
                dr.remove(m_compiler.getRoot());
                throw rce;

            } catch (JessException je) {
                exception = je;
            }

            m_rules.put(dr.getName(), dr);
            broadcastEvent(JessEvent.DEFRULE, dr, null);

            if (exception != null)
                throw exception;
            else
                return dr;
        }
    }

    /**
     * Remove a deffacts from the engine. If the given name is "*", remove all deffacts.
     * @param name the name of an existing deffacts construct, or "*"
     * @return the symbol TRUE
     */

    public final Value unDeffacts(String name) {
        if (name.equals("*"))
            m_deffacts.clear();
        else {
            name = resolveName(name);
            for (Iterator it = m_deffacts.iterator(); it.hasNext();) {
                Deffacts deffacts = (Deffacts) it.next();
                if (deffacts.getName().equals(name)) {
                    it.remove();
                    return Funcall.TRUE;
                }
            }
        }
        return Funcall.FALSE;
    }


    /**
     * Remove a rule or query from this Rete object. Removes all
     * subrules of the named rule as well.
     *
     * @param name the name of the rule or query
     * @return the symbol TRUE
     * @throws JessException if anything goes wrong
     */
    public final Value unDefrule(String name) throws JessException {
        synchronized (m_compiler) {
            HasLHS odr = findDefrule(name);
            if (odr != null) {
                m_rules.remove(resolveName(name));
                odr.remove(m_compiler.getRoot());
                HasLHS hlhs = odr;
                while (hlhs != null) {
                    if (hlhs instanceof Defrule)
                        for (Iterator e = m_agenda.listActivations(odr.getModule()); e.hasNext();) {
                            Activation a = (Activation) e.next();
                            if (a.getRule() == hlhs) {
                                removeActivation(a);
                            }
                        }
                    hlhs = hlhs.getNext();
                }
                commitActivations();
                broadcastEvent(JessEvent.DEFRULE | JessEvent.REMOVED, odr, null);
                return Funcall.TRUE;
            }
        }

        return Funcall.FALSE;
    }
    // **********************************************************************
    // Modules
    // **********************************************************************

    /**
     * Define a new module, which becomes current. The current module
     * is the one to which new constructs will be added.
     *
     * @param module a new Defmodule object
     */
    public void addDefmodule(Defmodule module)
            throws JessException {
        m_agenda.addDefmodule(module, this);
    }

    /**
     * Define a new module, which becomes current. The current module
     * is the one to which new constructs will be added.
     *
     * @deprecated As of Jess 7, use addDefmodule(Defmodule).
     */
    public void addDefmodule(String name)
            throws JessException {
        m_agenda.addDefmodule(new Defmodule(name, ""), this);
    }

    /**
     * Define a new module, which becomes current. The current module
     * is the one to which new constructs will be added.
     *
     * @deprecated As of Jess 7, use addDefmodule(Defmodule).
     */
    public void addDefmodule(String name, String comment)
            throws JessException {
        m_agenda.addDefmodule(new Defmodule(name, comment), this);
    }

    /**
     * Return the name of the current module. The current module is
     * the one to which new constructs will be added.
     */
    public String getCurrentModule() {
        return m_agenda.getCurrentModule();
    }

    /**
     * Change the current module. The current module is the one to
     * which new constructs will be added.
     *
     * @param name the name of the new current module
     * @return the old current module
     * @throws JessException if the name is not a module
     */
    public String setCurrentModule(String name) throws JessException {
        return m_agenda.setCurrentModule(name);
    }

    /**
     * List all modules.
     *
     * @return an iterator over the names of all modules
     */
    public Iterator listModules() throws JessException {
        return m_agenda.listModules();
    }

    /**
     * Query the focus module. The focus module is the one from which
     * rules are being fired.
     *
     * @return the name of the current focus module
     */
    public String getFocus() {
        return m_agenda.getFocus();
    }

    /**
     * Change the focus module. The focus module is the one from which
     * rules are being fired.
     *
     * @param name the module that should get focus
     */
    public void setFocus(String name) throws JessException {
        m_agenda.setFocus(name, this);
    }

    /**
     * Iterate over the module focus stack, from bottom to top. The
     * current focus module is returned last. The focus module is the
     * one from which rules are being fired.
     *
     * @return the iterator
     */
    public Iterator listFocusStack() throws JessException {
        return m_agenda.listFocusStack();
    }

    /**
     * Empty the module focus stack. The focus module is the one from
     * which rules are being fired.
     */
    public void clearFocusStack() {
        m_agenda.clearFocusStack();
    }

    /**
     * Remove the top module from the focus stack, and return it. If
     * expected in non-null, then this is a no-op unless expected
     * names the top module on the stack. The focus module is the one
     * from which rules are being fired.
     *
     * @param expected the expected module, or null
     * @return the name of the module on top the focus stack
     */
    public String popFocus(String expected) throws JessException {
        return m_agenda.popFocus(this, expected);
    }

    /**
     * Throw an exception if the argument isn't the name of a module.
     *
     * @param moduleName a proposed module name
     */

    public void verifyModule(String moduleName) throws JessException {
        m_agenda.verifyModule(moduleName);
    }

    /**
     * Decorate the name with the current module name, if it doesn't
     * already contain a module name.
     *
     * @param name a module name
     */

    public String resolveName(String name) {
        return m_agenda.resolveName(name);
    }


    // **********************************************************************
    // Dealing with the agenda: running, stopping, salience, etc.
    // **********************************************************************


    /**
     * Present all the facts on the agenda to a single Node.
     */
    void updateNodes(Hashtable n) throws JessException {
        m_factList.updateNodes(this, n);
    }

    /**
     * Info about a rule to fire.
     */
    void addActivation(Activation a) throws JessException {
        broadcastEvent(JessEvent.ACTIVATION, a, null);
        m_agenda.addActivation(a);
    }

    /**
     * An activation has been cancelled or fired; forget it
     */

    void removeActivation(Activation a) throws JessException {
        broadcastEvent(JessEvent.ACTIVATION | JessEvent.REMOVED, a, null);
        m_agenda.removeActivation(a);
    }

    void commitActivations() throws JessException {
        m_agenda.commitActivations(this);
    }


    /**
     * Return an Iterator over all the activiations for the current
     * module. Note that some of the activations may be cancelled or
     * already fired; check the return value of "isInactive()"for each
     * one to be sure. This is an expensive operation.
     *
     * @return an iterator over the agenda for the current module
     */
    public Iterator listActivations() {
        return m_agenda.listActivations();
    }

    /**
     * Return an Iterator over all the activiations for the named
     * module. Asterisk ("*") is not accepted. Note that some of the
     * activations may be cancelled or already fired; check the return
     * value of "isInactive()" for each one to be sure. This is an
     * expensive operation.
     *
     * @param moduleName the name of a module
     * @return an iterator over the agenda for the named module
     * @throws JessException if the module name is invalid
     * @see jess.Activation#isInactive
     */
    public Iterator listActivations(String moduleName) throws JessException {
        return m_agenda.listActivations(moduleName);
    }

    /**
     * Return the next activation record from the agenda. It's
     * possible that this rule will be wrong if something happens to
     * change the agenda before this rule fires.
     *
     * @return the next activation that will fire
     * @throws JessException if anything goes wrong
     */
    public Activation peekNextActivation() throws JessException {
        return m_agenda.peekNextActivation(this);
    }

    /**
     * The monitor of the object returned from this method will be signalled
     * whenever an activation appears. Thus a run-loop could wait on
     * this monitor when idle.
     *
     * @return the activation lock
     * @see #waitForActivations
     */

    public Object getActivationSemaphore() {
        return m_agenda.getActivationSemaphore();
    }

    /**
     * Waits on the activation lock as long as the agenda is
     * empty. Will return as soon as a rule is actvated. Can be called
     * in a run-loop to wait for more rules to fire.
     *
     * @see #getActivationSemaphore
     */

    public void waitForActivations() {
        m_agenda.waitForActivations();
    }

    /**
     * Tell this engine to use the given Strategy object to order the
     * rules on the agenda. You can implement your own strategies.
     *
     * @param s the new conflict resolution strategy
     * @return the name of the previous conflict resolution strategy
     * @throws JessException
     */
    public String setStrategy(Strategy s) throws JessException {
        return m_agenda.setStrategy(s, this);
    }


    /**
     * Retrieve the Strategy object this engine is using to order activations
     * on the agenda.
     *
     * @return the current conflict resolution strategy
     */
    public Strategy getStrategy() {
        return m_agenda.getStrategy();
    }

    /**
     * Set the salience evaluation behaviour. The behaviour can be one
     * of INSTALL, ACTIVATE, or EVERY_TIME; the default is INSTALL. When
     * the behaviour is INSTALL, a rule's salience is evulated once when
     * the rule is compiled. If it is ACTIVATE, it is computed each time
     * the rule is activated. If it is EVERY_TIME, salience evaluations
     * are done for all rules each time the next rule on the agenda is
     * to be chosen.
     *
     * @param method One of the acceptable values
     * @throws JessException If something goes wrong
     */

    public final void setEvalSalience(int method) throws JessException {
        m_agenda.setEvalSalience(method);
    }

    /**
     * Fetch the salience evaluation behaviour, which helps determine
     * the priority of rule firing. Returns one of the constants
     * INSTALL, ACTIVATE, or EVERY_TIME.
     *
     * @return the salience evaluation behaviour
     * @see #setEvalSalience
     */
    final public int getEvalSalience() {
        return m_agenda.getEvalSalience();
    }

    /**
     * Run the rule engine. When the engine is running, rules are
     * continuously removed from the agenda and fired in priority
     * order; the order is determined by the current conflict
     * resolution strategy.
     *
     * @return the actual number of rules fired
     * @throws JessException if anything goes wrong
     */
    public int run() throws JessException {
        broadcastEvent(JessEvent.RUN, this, null);
        return m_agenda.run(this);
    }

    /**
     * Allows a subclass to be notified immediately before a rule
     * fires. Subclasses can override this method to receive this
     * notification.
     *
     * @param a the activation record for the about-to-fire rule
     */
    protected void aboutToFire(Activation a) {
    }

    /**
     * Allows a subclass to be notified immediately after a rule
     * fires. Subclasses can override this method to receive this
     * notification.
     *
     * @param a the activation record for the just-fired rule
     */
    protected void justFired(Activation a) {
    }

    /**
     * Run the rule engine. If the number of rules fired becomes
     * greater than the argument, the engine stops. When the engine is
     * running, rules are continuously removed from the agenda and
     * fired in priority order; the order is determined by the current
     * conflict resolution strategy.
     *
     * @param max the maximum number of rules to fire
     * @return the number of rules that fired
     * @throws JessException if anything goes wrong
     */
    public int run(int max) throws JessException {
        broadcastEvent(JessEvent.RUN, this, null);
        return m_agenda.run(max, this);
    }

    /**
     * Run the rule engine until halt() is called. When no rules are
     * active, the calling Thread will be waiting on the activation
     * semaphore.  When the engine is running, rules are continuously
     * removed from the agenda and fired in priority order; the order
     * is determined by the current conflict resolution strategy.
     *
     * @return The number of rules that fired.
     * @throws JessException if an error occurs
     */

    public int runUntilHalt() throws JessException {
        broadcastEvent(JessEvent.RUN, this, null);
        return m_agenda.runUntilHalt(this);
    }

    /**
     * Stop the engine from firing rules.
     *
     * @throws JessException if an error occurs during event propagation
     */

    public void halt() throws JessException {
        broadcastEvent(JessEvent.HALT, this, null);
        m_agenda.halt();
    }

    /**
     * Returns true if halt has been called since the last run() call.
     *
     * @return true if the engine is halted
     */

    public boolean isHalted() {
        return m_agenda.isHalted();
    }


    /**
     * Find out the name of the currently firing rule.
     *
     * @return The name of the rule that is currently firing, if this is
     *         called while a rule is firing; otherwise, returns null.
     * @see #getThisActivation
     */
    public String getThisRuleName() {
        Activation a = getThisActivation();
        if (a != null)
            return a.getRule().getName();
        else
            return null;
    }

    /**
     * Get the activation record for the currently firing rule. An
     * activation record contains a Defrule and the list of facts that
     * made the rule active.
     *
     * @return the activation record for the rule that is currently
     *         firing, if this is called while a rule is firing; otherwise,
     *         returns null.
     * @see #getThisRuleName
     */
    public Activation getThisActivation() {
        return m_agenda.getThisActivation();
    }


    // **********************************************************************
    // Events and event listener support
    // **********************************************************************

    public void addJessListener(JessListener jel) {
        m_jes.addJessListener(jel);
    }

    public void removeJessListener(JessListener jel) {
        m_jes.removeJessListener(jel);
    }

    public Iterator listJessListeners() {
        return m_jes.listJessListeners();
    }

    /**
     * Query the current value of the event mask.
     * @return the event mask
     * @see #setEventMask
     */
    public int getEventMask() {
        return m_jes.getEventMask();
    }

    /**
     *
     * You can control which events this object will fire
     * using this method. The argument is the result
     * of logical-OR-ing together some of the constants in the
     * {@link jess.JessEvent} class. By default, the event mask is 0 and no
     * events are sent.
     * @param mask the new event mask
     * @see #getEventMask
     */
    public void setEventMask(int mask) {
        m_jes.setEventMask(mask);
    }

    final void broadcastEvent(int type, Object data, Context context) throws JessException {
        m_jes.broadcastDebugEvent(this, type, data, context);
        m_jes.broadcastEvent(this, type, data, context);
    }

    // **********************************************************************
    // Debug mode
    // **********************************************************************

    /**
     * Return whether the engine is in passive mode. In passive mode,
     * the engine won't send added rules to the compiler. This is used
     * by the JessDE implementation, but isn't useful to the vast
     * majority of clients.
     *
     * @param passiveMode the desired value of this property
     */
    public void setPassiveMode(boolean passiveMode) {
        m_passiveMode = passiveMode;
    }


    /**
     * Return whether the engine is in debug mode. In debug mode, the
     * engine can single-step. Special JessEvents are used for this
     * purpose.  This is used by the JessDE implementation, but isn't
     * useful to the vast majority of clients.
     *
     * @return whether the engine is in debug mode
     */
    public boolean isDebug() {
        return m_jes.isDebug();
    }

    /**
     * Specify whether the engine should be in debug mode. In debug
     * mode, the engine can single-step. Special JessEvents are used
     * for this purpose.  This is used by the JessDE implementation,
     * but isn't useful to the vast majority of clients.
     *
     * @param debug the desired state
     */
    public void setDebug(boolean debug) {
        m_jes.setDebug(debug);
    }

    /**
     * Note that a function is being called. This is used by the
     * JessDE implementation, but isn't useful to the vast majority of
     * clients.
     *
     * @param fc       the funcall being invoked
     * @param fileName the source file name
     * @param lineno   the source file line number
     */
    public static void recordFunction(Funcall fc, String fileName, int lineno) {
        recordFunction(fc, new LineNumberRecord(fileName, lineno));
    }

    /**
     * Note that a function is being called. This is used by the
     * JessDE implementation, but isn't useful to the vast majority of
     * clients.
     *
     * @param fc  the funcall being invoked
     * @param lnr the source file name and line number
     */
    public static void recordFunction(Funcall fc, LineNumberRecord lnr) {
        if (s_lineNumberTable == null)
            s_lineNumberTable = new IdentityHashMap();
        s_lineNumberTable.put(fc, lnr);
    }

    /**
     * Return information about the source location of a given
     * function call, if known. This method only works when the engine
     * is in debug mode. This is used by the JessDE implementation,
     * but isn't useful to the vast majority of clients.
     *
     * @param fc the funcall
     * @return source file and line number information, if known, or null
     */
    public static LineNumberRecord lookupFunction(Funcall fc) {
        if (s_lineNumberTable != null)
            return (LineNumberRecord) s_lineNumberTable.get(fc);
        else
            return null;
    }

    /**
     * Remove a listener so that it will no longer receive debug
     * events. This is used by the JessDE implementation, but isn't
     * useful to the vast majority of clients.
     *
     * @param jel the listener to remove
     */
    public void removeDebugListener(JessListener jel) {
        m_jes.removeDebugListener(jel);
    }

    /**
     * Add a listener so that it will receive debug events. This is
     * used by the JessDE implementation, but isn't useful to the vast
     * majority of clients.
     *
     * @param jel the listener to remove
     */
    public void addDebugListener(JessListener jel) {
        m_jes.addDebugListener(jel);
    }

    /**
     * Return an iterator over all listeners that will receive debug
     * events. This is used by the JessDE implementation, but isn't
     * useful to the vast majority of clients.
     *
     * @return the iterator
     */

    public Iterator listDebugListeners() {
        return m_jes.listDebugListeners();
    }

    // **********************************************************************
    // Bloading and serialization
    // **********************************************************************

    private void readObject(ObjectInputStream stream)
            throws IOException, ClassNotFoundException {
        stream.defaultReadObject();
        m_routers = new Routers();
        m_tis = new TextReader(true);
        m_jesp = initInternalParser();
        m_jes = new JessEventSupport(this);
        m_definstanceList.setEngine(this);
        m_globalContext.setEngine(this);
        m_classSource.setEngine(this);
    }

    /**
     * Read this object's state from the given stream. Similar to the
     * readObject() method used in Serialization, except it loads new
     * state into an existing object. The data on the InputStream must
     * have been written by a call to bsave().
     *
     * @param is an InputStream
     */

    public void bload(InputStream is)
            throws IOException, ClassNotFoundException {
        ObjectInputStream ois = new ObjectInputStream(is);

        m_globalContext = (Context) ois.readObject();
        m_globalContext.setEngine(this);
        m_resetGlobals = ois.readBoolean();
        m_deffacts = (java.util.List) ois.readObject();
        m_defglobals = (java.util.List) ois.readObject();
        m_functions = (Map) ois.readObject();
        m_factList = (FactList) ois.readObject();
        m_definstanceList = (DefinstanceList) ois.readObject();
        m_definstanceList.setEngine(this);
        m_rules = (Map) ois.readObject();
        m_compiler = (ReteCompiler) ois.readObject();
        m_storage = (Map) ois.readObject();
        m_agenda = (Agenda) ois.readObject();
        m_classSource = (ClassSource) ois.readObject();
        m_classSource.setEngine(this);
        m_watchInfo = (boolean[]) ois.readObject();
        m_features = (HashSet) ois.readObject();
    }

    /**
     * Save this object's state out to the given stream. Similar to
     * the writeObject() method used in serialization, but only writes
     * the state of this object, not the object itself. You can read
     * this into another Rete object using the bload() method.
     *
     * @param os an OutputStream
     */

    public void bsave(OutputStream os) throws IOException {
        ObjectOutputStream oos = new ObjectOutputStream(os);

        oos.writeObject(m_globalContext);
        oos.writeBoolean(m_resetGlobals);
        oos.writeObject(m_deffacts);
        oos.writeObject(m_defglobals);
        oos.writeObject(m_functions);
        oos.writeObject(m_factList);
        oos.writeObject(m_definstanceList);
        oos.writeObject(m_rules);
        oos.writeObject(m_compiler);
        oos.writeObject(m_storage);
        oos.writeObject(m_agenda);
        oos.writeObject(m_classSource);
        oos.writeObject(m_watchInfo);
        oos.writeObject(m_features);
        oos.flush();
    }

    // **********************************************************************
    // Defclass and definstance support
    // **********************************************************************

    /**
     * Make a shadow fact for the given object and add it to working
     * memory. The boolean argument indicates whether or not to
     * install a PropertyChangeListener to listen for change
     * events. Uses the global context, which is usually what you
     * want.
     *
     * @param jessTypename the name of a defclass
     * @param object       a Java object of the defclass's type
     * @param dynamic      true if PropertyChangeListeners should be used
     * @return a FactIdValue containing the shadow fact for the object
     * @throws JessException if anything goes wrong
     */
    public Value definstance(String jessTypename, Object object,
                             boolean dynamic) throws JessException {
        return definstance(jessTypename, object, dynamic, getGlobalContext());
    }

    /**
     * Make a shadow fact for the given object and add it to working
     * memory. The boolean argument indicates whether or not to
     * install a PropertyChangeListener to listen for change events.
     *
     * @param jessTypename the name of a defclass
     * @param object       a Java object of the defclass's type
     * @param dynamic      true if PropertyChangeListeners should be used
     * @param context      used for asserting the fact
     * @return a FactIdValue containing the shadow fact for the object
     * @throws JessException if anything goes wrong
     */
    public Value definstance(String jessTypename, Object object,
                             boolean dynamic, Context context)
            throws JessException {
        broadcastEvent(JessEvent.DEFINSTANCE, object, context);
        jessTypename = resolveTemplateNameInFocusModule(jessTypename);
        return m_definstanceList.definstance(resolveTemplateNameInFocusModule(jessTypename), object, dynamic, context);
    }

    String resolveTemplateNameInFocusModule(String jessTypename) throws JessException {
        if (jessTypename.indexOf("::") != -1)
            return jessTypename;
        else if (getFocus().equals(getCurrentModule()))
            return jessTypename;

        String focusName = getFocus() + "::" + jessTypename;
        if (findDeftemplate(focusName) != null)
            return focusName;
        else
            return jessTypename;
    }

    /**
     * Remove the given object from working memory. Tell this engine
     * to stop pattern matching on the given object, and retract the
     * shadow fact.
     *
     * @param object an object currently represented in working memory.
     * @return the shadow fact for the object
     * @throws JessException if object isn't a definstanced object, or on error
     */

    public Fact undefinstance(Object object) throws JessException {

        Fact f = m_definstanceList.undefinstance(object);
        broadcastEvent(JessEvent.DEFINSTANCE | JessEvent.REMOVED, object, null);
        return f;
    }

    void undefinstanceNoRetract(Object object) throws JessException {

        m_definstanceList.undefinstanceNoRetract(object);
        broadcastEvent(JessEvent.DEFINSTANCE | JessEvent.REMOVED, object, null);
    }

    /**
     * Bring a shadow fact up to date. If the properties of the given
     * object, assumed to be a definstanced object, have changed, its
     * corresponding shadow fact will be updated.
     *
     * @param object a previously definstanced object
     * @return the shadow fact, as a FactIDValue
     * @throws JessException if object isn't a definstanced object, or on error
     */
    public Value updateObject(Object object) throws JessException {
        return m_definstanceList.updateObject(object);
    }

    /**
     * Bring one slot of a shadow fact up to date. If the named property of the given
     * object, assumed to be a definstanced object, has changed, its
     * corresponding shadow fact will be updated.
     *
     * @param object a previously definstanced object
     * @param slotName the name of a property of that object
     * @return the shadow fact, as a FactIDValue
     * @throws JessException if object isn't a definstanced object, or on error
     */
    public Value updateObject(Object object, String slotName) throws JessException {
        return m_definstanceList.updateObject(object, slotName);
    }

    /**
     * Add an object to working memory using various defaults. In this implementation, Jess
     * assumes you want to use an existing deftemplate with the same name as the class of this
     * object, minus the package names. If no such deftemplate exists, one is created. The object
     * is added without support for PropertyChangeEvents.
     *
     * @param o an object to add to working memory
     * @throws JessException if anything goes wrong
     */
    public void add(Object o) throws JessException {
        String name = o.getClass().getName();
        // TODO Smart search through defclasses
        String shortName = ClassSource.classNameOnly(name);
        if (findDeftemplate(shortName) == null)
            defclass(shortName, name, null);
        definstance(shortName, o, false);
    }

    /**
     * Add a collection of objects to working memory. Calls add(Object) on each item in the collection.
     *
     * @see Rete#add(Object)
     * @param c a collection of objects to be added to working memory
     * @throws JessException if anything goes wrong
     */

    public void addAll(Collection c) throws JessException {
        addAll(c.iterator());
    }

    /**
     * Add a collection of objects to working memory. Calls add(Object) on each item in the collection.
     *
     * @see Rete#add(Object)
     * @param it Iterator over a collection of objects to be added to working memory
     * @throws JessException if anything goes wrong
     */

    public void addAll(Iterator it) throws JessException {
        while (it.hasNext())
            add(it.next());
    }

    /**
     * Add a defclass definition (a deftemplate) to this engine. One
     * slot is included for each JavaBeans property of the class. In
     * addition, slots representing public member variables can be
     * included.
     *
     * @param jessName               the name Jess should use for this defclass
     * @param clazz                  the name of the Java class
     * @param parent                 if non-null, a parent deftemplate or defclass name
     * @param includeMemberVariables if true, slots corresponding to
     *                               public member variables are created.
     */

    public Value defclass(String jessName, String clazz, String parent,
                          boolean includeMemberVariables)
            throws JessException {
        broadcastEvent(JessEvent.DEFCLASS, jessName, null);
        return m_definstanceList.defclass(jessName, clazz, parent,
                                          includeMemberVariables);
    }

    /**
     * Add a defclass definition (a deftemplate) to this engine. One
     * slot will be created for each JavaBeans property of the class.
     *
     * @param jessName The name Jess should use for this defclass
     * @param clazz    The name of the Java class
     * @param parent   If non-null, a parent deftemplate or defclass name
     */

    public Value defclass(String jessName, String clazz, String parent)
            throws JessException {
        broadcastEvent(JessEvent.DEFCLASS, jessName, null);
        return m_definstanceList.defclass(jessName, clazz, parent, false);
    }

    FactList getFactList() {
        return m_factList;
    }

    DefinstanceList getDefinstanceList() {
        return m_definstanceList;
    }

    // **********************************************************************
    // Miscellaneous
    // **********************************************************************

    /**
     * Fetch the global execution context. This gives you access to
     * defglobals and variables defined at the global scope.
     *
     * @return The global execution context.
     */
    public final Context getGlobalContext() {
        return m_globalContext;
    }

    /**
     Evaluate a Jess expression in this engine's global context. The
     result is returned as a jess.Value; you can use the methods of
     that class to extract the data. For example, here we define a
     deffunction and then call it, storing the result in an int variable:

     <pre>
     Rete r = new Rete();
     r.eval("(deffunction square (?n) (return (* ?n ?n)))");
     Value v = r.eval("(square 3)");
     int nine = v.intValue(r.getGlobalContext()));
     </pre>

     <p> Note that you may only pass one function call or construct at a
     time to <tt>eval().</tt></p>

     @param cmd a string containing a Jess expression
     @return the result of evaluating the expression
     @throws JessException if anything goes wrong
     */

    public Value eval(String cmd) throws JessException {
        return eval(cmd, m_globalContext);
    }


    /**
     * Evaluate a Jess expression in this engine's global context. The
     * result is returned as a jess.Value; you can use the methods of
     * that class to extract the data.
     *
     * @deprecated Use eval() instead.
     * @see Rete#eval(String)
     * @param cmd a string containing a Jess expression
     * @return the result of evaluating the expression
     * @throws JessException if anything goes wrong
     */
    public Value executeCommand(String cmd) throws JessException {
        return eval(cmd, m_globalContext);
    }

    /**
     Evaluate a Jess expression in the given execution context. The
     result is returned as a jess.Value; you can use the methods of
     that class to extract the data. For example, here we define a
     deffunction and then call it, storing the result in an int variable:

     <pre>
     Rete r = new Rete();
     r.eval("(deffunction square (?n) (return (* ?n ?n)))");
     Value v = r.eval("(square 3)");
     int nine = v.intValue(r.getGlobalContext()));
     </pre>

     <p> Note that you may only pass one function call or construct at a
     time to <tt>eval().</tt></p>

     @param cmd a string containing a Jess expression
     @param context the evaluation context
     @return the result of evaluating the expression
     @throws JessException if anything goes wrong
     */
    public Value eval(String cmd, Context context) throws JessException {
        synchronized (m_tis) {
            m_tis.clear();
            m_jesp.clear();
            m_tis.appendText(cmd);
            return m_jesp.parse(false, context);
        }
    }

    /**
     * Evaluate a Jess expression in the given context. The result is
     * returned as a jess.Value; you can use the methods of that class
     * to extract the data.
     *
     * @deprecated Use eval() instead.
     * @see Rete#eval(String)
     * @param cmd a string containing a Jess expression
     * @param context the evaluation context
     * @return the result of evaluating the expression
     * @throws JessException if anything goes wrong
     */
    public Value executeCommand(String cmd, Context context) throws JessException {
        synchronized (m_tis) {
            m_tis.clear();
            m_jesp.clear();
            m_tis.appendText(cmd);
            return m_jesp.parse(false, context);
        }
    }

    private Jesp initInternalParser() {
        Jesp jesp = new Jesp(m_tis, this);
        jesp.setFileName("<eval pipe>");
        return jesp;
    }

    /**
     * Set the resetGlobals property. When resetGlobals is true, the
     * initializers of defglobals are evaluated when (reset) is
     * executed. When it is false, (reset) doesn't reinitialize
     * defglobals.
     *
     * @param reset The value of this property
     */
    public final void setResetGlobals(boolean reset) {
        m_resetGlobals = reset;
    }

    /**
     * Return the resetGlobals property. When resetGlobals is true,
     * the initializers of defglobals are evaluated when (reset) is
     * executed. When it is false, (reset) doesn't reinitialize
     * defglobals.
     *
     * @return The value of this property
     */
    final public boolean getResetGlobals() {
        return m_resetGlobals;
    }

    /**
     * Fetch the ReteCompiler object used by the engine. You
     * probabably shouldn't use this for anything!
     *
     * @return the Compiler object
     */

    final ReteCompiler getCompiler() {
        return m_compiler;
    }

    /**
     * Store a value in the engine under a given name for later
     * retrieval. The store and fetch methods provide a simple way to
     * exchange data between Jess and Java. They give you easy access
     * to a single HashMap from both languages, so you can use the
     * HashMap as a "mailbox". Use store() or the Jess function
     * (store) to save a value in this map; use fetch() or the Jess
     * function (fetch) to retrieve the value.
     *
     * @param name a key under which to save the value
     * @param val  the value to store
     * @return any old value stored under this name, or null.
     * @see Rete#fetch
     */

    public Value store(String name, Value val) {
        if (val == null)
            return (Value) m_storage.remove(name);
        else
            return (Value) m_storage.put(name, val);
    }

    /**
     * Store a value in the engine under a given name for later
     * retrieval by fetch. The Object is first wrapped in a new
     * jess.Value object. The store and fetch methods provide a simple
     * way to exchange data between Jess and Java. They give you easy
     * access to a single HashMap from both languages, so you can use
     * the HashMap as a "mailbox". Use store() or the Jess function
     * (store) to save a value in this map; use fetch() or the Jess
     * function (fetch) to retrieve the value.
     *
     * @param name a key under which to file the value
     * @param val  the value to store
     * @return any old value stored under this name, or null
     * @see Rete#fetch
     */

    public Value store(String name, Object val) {
        if (val == null)
            return (Value) m_storage.remove(name);
        else
            return (Value) m_storage.put(name, new Value(val));
    }

    /**
     * Retrieve an object previously stored with store(). The store
     * and fetch methods provide a simple way to exchange data between
     * Jess and Java. They give you easy access to a single HashMap
     * from both languages, so you can use the HashMap as a
     * "mailbox". Use store() or the Jess function (store) to save a
     * value in this map; use fetch() or the Jess function (fetch) to
     * retrieve the value.<p/>
     * <p/>
     * The result is returned as a jess.Value * object; you must use
     * the methods of that class to extract the * underlying data.
     *
     * @param name the key under which to find an object
     * @return the object, or null if not found.
     * @see Rete#store
     */

    public Value fetch(String name) {
        return (Value) m_storage.get(name);
    }

    /**
     * Clear the storage used by store() and fetch().
     */

    public void clearStorage() {
        m_storage.clear();
    }


    int getTime() {
        return m_factList.getTime();
    }


    /**
     * Return the Factory Jess will use to create Rete Tokens. This
     * method allows extensions like FuzzyJess to use special
     * augmented tokens to store additional information. The vast
     * majority of clients will never use this method.
     *
     * @return the current Token Factory implementation
     */
    public static Factory getFactory() {
        return s_factory;
    }

    /**
     * Set the Factory Jess will use to create Rete Tokens. This
     * method allows extensions like FuzzyJess to use special
     * augmented tokens to store additional information. The vast
     * majority of clients will never use this method.
     *
     * @param f a jess.Factory implementation
     */
    public static void setFactory(Factory f) {
        s_factory = f;
    }

    private void loadScriptlib() {
        try {
            batch(LIBRARY_NAME);
        } catch (JessException je) {
            // TODO Explain the problem!
            je.printStackTrace();
        }
    }

    /**
     * Execute a file of Jess language code or JessML code. Jess will
     * look on the file system relative to the current directory, and
     * also along the CLASSPATH, to find the named file. If Jess is
     * running in an applet, the document base will also be checked.
     *
     * @param filename the name of a file or resource
     * @return the value of the last item parsed from the file, or FALSE
     * @throws JessException if there are I/O problems, or syntax errors in the file
     * @see Batch#batch
     */
    public Value batch(String filename) throws JessException {
        return Batch.batch(filename, this);
    }

    /**
     * Invoke the named query and return the results. The QueryResult
     * has an API similar to java.sql.ResultSet. The global context is
     * used to resolve variables.
     *
     * @param name   the name of the query
     * @param params the query's parameters
     * @return a QueryResult containing all the matches
     * @throws JessException if anything goes wrong
     */

    public QueryResult runQueryStar(String name, ValueVector params)
            throws JessException {
        return runQueryStar(name, params, getGlobalContext());
    }


    /**
     * Invoke the named query and return the results. The QueryResult
     * has an API similar to java.sql.ResultSet. The given context is
     * used to resolve variables.
     *
     * @param name   the name of the query
     * @param params the query's parameters
     * @return a QueryResult containing all the matches
     * @throws JessException if anything goes wrong
     */

    public QueryResult runQueryStar(String name, ValueVector params, Context context)
            throws JessException {
        Iterator it = doRunQuery(name, params, context);
        return new QueryResult(it, (Defquery) findDefrule(name), context);
    }

    /**
     * Invoke the named query and return an iterator over the matching
     * results. Each match is a jess.Token object. The global context
     * is used to resolve variables.
     *
     * @param name   the name of the query
     * @param params the query's parameters
     * @return an iterator over all results
     * @throws JessException if anything goes wrong
     * @deprecated Since Jess 7.0, superceded by runQueryStar()
     */

    public Iterator runQuery(String name, ValueVector params)
            throws JessException {
        return doRunQuery(name, params, getGlobalContext());
    }

    /**
     * Invoke the named query and return an iterator over the matching
     * results. Each match is a jess.Token object. The global context
     * is used to resolve variables.
     *
     * @param name   the name of the query
     * @param params the query's parameters
     * @return an iterator over all results
     * @throws JessException if anything goes wrong
     * @deprecated Since Jess 7.0, superceded by runQueryStar()
     */

    public Iterator runQuery(String name, ValueVector params, Context context)
            throws JessException {
        return doRunQuery(name, params, context);
    }

    Iterator doRunQuery(String name, ValueVector params, Context context) throws JessException {
        HasLHS lhs = findDefrule(name);
        if (lhs == null || !(lhs instanceof Defquery))
            throw new JessException("runQuery", "No such query:", name);

        Defquery query = (Defquery) lhs;

        if ((params.size()) != query.getNVariables())
            throw new JessException("runQuery", "Wrong number of variables for query", name);

        // Create the query-trigger fact
        Fact f = new Fact(query.getQueryTriggerName(), this);
        f.setSlotValue(RU.DEFAULT_SLOT_NAME, new Value(params, RU.LIST));

        // Assert the fact, blocking access to other queries; then return the
        // results, which clears the query
        synchronized (getCompiler()) {
            synchronized (query) {
                query.clearResults();
                assertFact(f, context);
                // Allow backwards chaining to occur
                if (query.getMaxBackgroundRules() > 0)
                    run(query.getMaxBackgroundRules());
                Iterator it = query.getResults();

                query.clearResults();
                retract(f);
                return it;
            }
        }
    }

    /**
     * Invoke the named query and return the count of matching
     * results. The global context is used to resolve variables.
     *
     * @param name   the name of the query
     * @param params the query's parameters
     * @return the count of matching results
     * @throws JessException if anything goes wrong
     */

    public int countQueryResults(String name, ValueVector params)
            throws JessException {
        return countQueryResults(name, params, getGlobalContext());
    }

    /**
     * Invoke the named query and return the count of matching
     * results. The given context is used to resolve variables.
     *
     * @param name    the name of the query
     * @param params  the query's parameters
     * @param context an execution context
     * @return the count of matching results
     * @throws JessException if anything goes wrong
     */

    public int countQueryResults(String name, ValueVector params, Context context) throws JessException {
        HasLHS lhs = findDefrule(name);
        if (lhs == null || !(lhs instanceof Defquery))
            throw new JessException("countQueryResults", "No such query:", name);

        Defquery query = (Defquery) lhs;

        if ((params.size()) != query.getNVariables())
            throw new JessException("countQueryResults", "Wrong number of variables for query", name);

        // Create the query-trigger fact
        Fact f = new Fact(query.getQueryTriggerName(), this);
        f.setSlotValue(RU.DEFAULT_SLOT_NAME, new Value(params, RU.LIST));

        // Assert the fact, blocking access to other queries; then return the
        // results, which clears the query
        synchronized (getCompiler()) {
            synchronized (query) {
                query.clearResults();
                assertFact(f, context);
                // Allow backwards chaining to occur
                if (query.getMaxBackgroundRules() > 0)
                    run(query.getMaxBackgroundRules());
                int count = query.countResults();

                query.clearResults();
                retract(f);
                return count;
            }
        }
    }


    // **********************************************************************
    // The watch facility
    // **********************************************************************

    /**
     * Produce some debugging information. The argument specifies which
     * kind of event will be reported. The output goes to the "WSTDOUT" router.
     *
     * @param which one of the constants in WatchConstants
     * @throws JessException if the argument is invalid
     * @see jess.WatchConstants
     */

    public void watch(int which) throws JessException {
        m_watchInfo[which] = true;
        int mask;
        switch (which) {
            case WatchConstants.RULES:
                mask = JessEvent.DEFRULE_FIRED;
                break;

            case WatchConstants.FACTS:
                mask = JessEvent.FACT;
                break;

            case WatchConstants.ACTIVATIONS:
                mask = JessEvent.ACTIVATION;
                break;

            case WatchConstants.COMPILATIONS:
                mask = JessEvent.DEFRULE;
                break;

            case WatchConstants.FOCUS:
                mask = JessEvent.FOCUS;
                break;

            default:
                throw new JessException("watch", "Bad argument ", which);
        }
        m_watchInfo[which] = true;
        mask = getEventMask() | mask | JessEvent.CLEAR;
        setEventMask(mask);
    }

    /**
     * Cancel some debugging information. The argument specifies which
     * kind of event will no longer be reported.
     *
     * @param which one of the constants in WatchConstants
     * @throws JessException if the argument is invalid
     * @see jess.WatchConstants
     */
    public void unwatch(int which) throws JessException {
        int mask;
        switch (which) {
            case WatchConstants.RULES:
                mask = JessEvent.DEFRULE_FIRED;
                break;

            case WatchConstants.FACTS:
                mask = JessEvent.FACT;
                break;

            case WatchConstants.ACTIVATIONS:
                mask = JessEvent.ACTIVATION;
                break;

            case WatchConstants.COMPILATIONS:
                mask = JessEvent.DEFRULE;
                break;

            case WatchConstants.FOCUS:
                mask = JessEvent.FOCUS;
                break;

            default:
                throw new JessException("unwatch", "Bad argument ", which);
        }
        m_watchInfo[which] = false;
        mask = getEventMask() & ~mask;
        setEventMask(mask);
    }

    /**
     * Produce all possible debugging info. Equivalent to calling watch()
     * multiple times using each legal argument in succession.
     */
    public void watchAll() {
        for (int i = 0; i < m_watchInfo.length; ++i)
            m_watchInfo[i] = true;

        int mask = JessEvent.DEFRULE |
                JessEvent.DEFRULE_FIRED |
                JessEvent.FACT |
                JessEvent.FOCUS |
                JessEvent.ACTIVATION;
        mask = getEventMask() | JessEvent.CLEAR | mask;
        setEventMask(mask);
    }

    /**
     * Cancel all debugging info. Equivalent to calling unwatch()
     * using each legal argument in succession.
     */
    public void unwatchAll() {
        for (int i = 0; i < m_watchInfo.length; ++i)
            m_watchInfo[i] = false;

        int mask = JessEvent.DEFRULE |
                JessEvent.DEFRULE_FIRED |
                JessEvent.FACT |
                JessEvent.FOCUS |
                JessEvent.ACTIVATION;
        mask = getEventMask() & ~mask;
        setEventMask(mask);
    }

    boolean watchingAny() {
        for (int i = 0; i < m_watchInfo.length; ++i)
            if (m_watchInfo[i])
                return true;
        return false;
    }

    boolean watching(int which) {
        return m_watchInfo[which];
    }

    /**
     * Responds to a JessEvent by emitting "watch" messages. This
     * method is just an implementation detail, part of the "watch"
     * facility. Each Rete object is registered with itself as a
     * listener.
     *
     * @param je An event object
     */
    public void eventHappened(JessEvent je) {
        if (!watchingAny())
            return;
        int type = je.getType();
        boolean remove = (type & JessEvent.REMOVED) != 0;
        boolean modified = (type & JessEvent.MODIFIED) != 0;

        PrintWriter pw = getWatchRouter();
        if (pw == null)
            return;

        switch (type & ~JessEvent.REMOVED & ~JessEvent.MODIFIED) {

            case JessEvent.FACT:
                {
                    if (watching(WatchConstants.FACTS)) {
                        Fact f = (Fact) je.getObject();
                        pw.print(remove ? " <== " : modified ? " <=> " : " ==> ");
                        pw.print("f-");
                        pw.print(f.getFactId());
                        pw.print(" ");
                        pw.println(f);
                        pw.flush();
                    }
                    break;
                }

            case JessEvent.FOCUS:
                {
                    if (watching(WatchConstants.FOCUS)) {
                        pw.print(remove ? " <== " : " ==> ");
                        pw.print("Focus ");
                        pw.println(je.getObject());
                        pw.flush();
                    }
                    break;
                }

            case JessEvent.DEFRULE_FIRED:
                {
                    if (watching(WatchConstants.RULES))
                        ((Activation) je.getObject()).debugPrint(getOutStream());
                    break;
                }

            case JessEvent.ACTIVATION:
                {
                    if (watching(WatchConstants.ACTIVATIONS)) {
                        Activation a = (Activation) je.getObject();
                        pw.print(remove ? "<== " : "==> ");
                        pw.print("Activation: ");
                        pw.print(a.getRule().getDisplayName());
                        pw.print(" : ");
                        pw.println(a.getToken().factList());
                        pw.flush();
                    }
                    break;
                }

            case JessEvent.DEFRULE:
                {
                    if (watching(WatchConstants.COMPILATIONS) && !remove) {
                        pw.println(((HasLHS) je.getObject()).getCompilationTrace());
                        pw.flush();
                    }
                    break;
                }

            default:
                break;
        }
    }

    /**
     * Sets the router the "watch" facility will use for output. The
     * default is WSTDOUT. The named router must exist.
     *
     * @param s the name of the router to use
     * @return the name of the previous watch router
     * @throws JessException if the router doesn't exist
     */
    public String setWatchRouter(String s) throws JessException {
        if (m_routers.getOutputRouter(s) == null)
            throw new JessException("Rete.setWatchRouter", "Invalid router name", s);
        String old = m_watchRouter;
        m_watchRouter = s;
        return old;
    }

    PrintWriter getWatchRouter() {
        return m_routers.getOutputRouterAsPrintWriter(m_watchRouter);
    }

    // ****************************************************************
    // Require/provide
    // ****************************************************************

    /**
     * Register a named feature, as if by the "provide"
     * function. After calling this method, calling isFeatureDefined
     * with the same argument will always return true.
     *
     * @param name the name of the feature
     */

    public void defineFeature(String name) {
        m_features.add(name);
    }

    /**
     * Return whether a feature name has been registered by the
     * "provide" function.
     *
     * @param name the name of the feature
     * @return true if the feature is defined
     */

    public boolean isFeatureDefined(String name) {
        return m_features.contains(name);
    }

    // ****************************************************************
    // Logical support facility
    // ****************************************************************

    void removeLogicalSupportFrom(Token token, Fact fact) {
        m_factList.removeLogicalSupportFrom(token, fact);
    }

    /**
     * Returns a list of one or more jess.Token objects that provide
     * logical support for this fact.  This method returns null if
     * there is no specific logical support. You can use the
     * Token.size() method to check how many supporting Facts are in a
     * Token, and the Token.fact() method to check each supporting
     * Fact in turn. This is a fast operation, taking O(ln N) time,
     * where N is the number of facts that have logical support.
     *
     * @param fact a fact of interest
     * @return a list of supporting Token objects, or null if there is
     *         unconditional support
     * @see jess.Token
     * @see jess.Token#fact
     * @see jess.Token#size
     */

    public List getSupportingTokens(Fact fact) {
        return m_factList.getSupportingTokens(fact);
    }

    /**
     * Returns a list of Fact objects that receive logical support
     * from the argument.  This method is potentially expensive, as it
     * takes time proportional to O(N), where N is the number of facts
     * currently receiving logical support of any kind.
     *
     * @param supporter a fact of interest
     * @return a list of zero or more Fact objects
     */

    public List getSupportedFacts(Fact supporter) {
        return m_factList.getSupportedFacts(supporter);
    }

    /**
     * If this Rete object is part of an applet, and the applet was
     * supplied as a constructor argument, then this method will
     * return it.
     *
     * @return the Applet this Rete object is a part of, if any
     */
    public Applet getApplet() {
        return m_classSource.getApplet();
    }

    /**
     * Return the Class that represents the "app object" for this
     * Rete. This class's ClassLoader plays a special role: it is used
     * to load classes named in Jess language code.
     *
     * @return the Class object whose ClassLoader will be used
     */
    public Class getAppObjectClass() {
        return m_classSource.getAppObjectClass();
    }

    /**
     * Notify this Rete object that it is part of an Applet. The
     * Applet becomes the "app object", and its ClassLoader will be
     * used to load classes named in Jess code.
     *
     * @param applet the Applet
     */

    public void setApplet(Applet applet) {
        m_classSource.setApplet(applet);
    }

    /**
     * Specify the "app object" for this Rete. The app object's
     * ClassLoader will be used to load classes named in Jess code. By
     * default, a Rete object is its own app object.
     *
     * @param appObject the "app object".
     */
    public void setAppObject(Object appObject) {
        m_classSource.setAppObject(appObject);
    }

    /**
     * Specify the ClassLoader that will be used to load classes named
     * in Jess code. By default, the context ClassLoader is used.
     *
     * @param loader the ClassLoader to use
     */
    public void setClassLoader(ClassLoader loader) {
        m_classSource.setClassLoader(loader);
    }

    /**
     * Return the ClassLoader that will be used to find classes named
     * in Jess ocde. By default, the context ClassLoader is used.
     *
     * @return the ClassLoader used to load user classes
     */

    public ClassLoader getClassLoader() {
        return m_classSource.getClassLoader();
    }

    /**
     * Load a class using the active ClassLoader. Previously loaded
     * classes are cached. The import tables are used, so the given
     * name doesn't have to be fully qualified.
     *
     * @param className the name of the class
     * @return the Class object
     * @throws ClassNotFoundException if the class could not be loaded
     */
    public Class findClass(String className) throws ClassNotFoundException {
        return m_classSource.findClass(className);
    }

    /**
     * Load a resource using the current ClassLoader's getResource() method.
     *
     * @param name the name of the resource
     * @return a URL for the resource
     */
    public URL getResource(String name) {
        return m_classSource.getResource(name);
    }

    /**
     * Make all the classes in a package available for use in
     * unqualified form from Jess code. The package name should
     * include a trailing "." -- for example, "java.lang." .
     *
     * @param pack the name of a Java package
     */

    public void importPackage(String pack) {
        m_classSource.importPackage(pack);
    }

    /**
     * Make a single class name available for use in unqualified form
     * in Jess code. Also for each public static method and variable,
     * create a Jess function with name of the form "Class.member"
     * which, when called, defers to the given static member, either
     * calling the method or reading the value of the variable.
     *
     * @param clazz the fully-qualified name of the class
     * @throws JessException if the class is not found
     */
    public void importClass(String clazz) throws JessException {
        m_classSource.importClass(clazz);
    }

    /**
     * Returns "[Rete]".
     *
     * @return the String "[Rete]".
     */
    public String toString() {
        return "[Rete]";
    }

}



