package jess;

import jess.xml.JessSAXParser;
import org.xml.sax.InputSource;

import java.io.*;
import java.net.URL;

/**
 * This class implements the "batch" command, and several static
 * methods are available that you can call to load Jess code.
 * <P>
 * (C) 2005 Sandia National Laboratories<br>
 * @see Rete#batch
 */

public class Batch implements Userfunction, Serializable {
    public String getName() {
        return "batch";
    }

    /**
     * Execute a file of Jess code, either in the Jess rule language or in JessML. The filename
     * will be looked for first on the file system; if it's not found, Jess will search the
     * classpath. If Jess is used in an Applet, the applet's document base will be checked first.
     * The code will be evaluated in the Rete object's global execution context. Any errors
     * will terminate parsing.
     *
     * @param filename the name of a file, or a relative or absolute path
     * @param engine the Rete object to load with the constructs from the file
     * @return the last expression parsed
     * @throws JessException if anything goes wrong
     */
    public static Value batch(String filename, Rete engine)
            throws JessException {
        return batch(filename, engine, engine.getGlobalContext());
    }

    /**
     * Execute a file of Jess code, either in the Jess rule language or in JessML. The filename
     * will be looked for first on the file system; if it's not found, Jess will search the
     * classpath. If Jess is used in an Applet, the applet's document base will be checked first.
     * The code will be evaluated in the given execution context. Any errors will terminate parsing.
     *
     * @param filename the name of a file, or a relative or absolute path
     * @param engine the Rete object to load with the constructs from the file
     * @param context the execution context to use
     * @return the last expression parsed
     * @throws JessException if anything goes wrong
     */
    public static Value batch(String filename, Rete engine, Context context)
            throws JessException {
        return batch(filename, engine, context, new ErrorHandler.DefaultHandler());
    }

    /**
     * Execute a file of Jess code, either in the Jess rule language or in JessML. The filename
     * will be looked for first on the file system; if it's not found, Jess will search the
     * classpath. If Jess is used in an Applet, the applet's document base will be checked first.
     * The code will be evaluated in the given execution context. If any errors are encountered, the
     * given error handler will be invoked. If the error handler rethrows the exception, parsing
     * will terminate; otherwise parsing will continue.
     *
     * @param filename the name of a file, or a relative or absolute path
     * @param engine the Rete object to load with the constructs from the file
     * @param context the execution context to use
     * @param handler notified in case of exceptions
     * @return the last expression parsed
     * @throws JessException if anything goes wrong
     */
    public static Value batch(String filename, Rete engine, Context context, ErrorHandler handler)
        throws JessException {

        Value v = Funcall.FALSE;
        PushbackReader fis = null;
        try {
            fis = findDocument(engine, filename);

            if (isXMLDocument(fis)) {
                parseXMLDocument(engine, fis, handler);
                v = Funcall.TRUE;
            } else {
                Jesp j = new Jesp(fis, engine);
                j.setFileName(filename);
                while (true) {
                    try {
                        v = j.parse(false, context);
                        break;
                    } catch (JessException ex) {
                        handler.handleError(ex);
                    }
                }
            }

        } catch (IOException ex) {
            handler.handleError(new JessException("batch", "I/O Exception", ex));

        } finally {
            if (fis != null) try { fis.close(); } catch (IOException ioe) {}
        }
        return v;
    }

    static void parseXMLDocument(Rete engine, PushbackReader fis, ErrorHandler handler) throws JessException {
        JessSAXParser parser = new JessSAXParser(engine);
        while (true) {
            try {
                parser.parse(new InputSource(fis));
                break;
            } catch (JessException ex) {
                handler.handleError(ex);
            }
        }
    }

    public static PushbackReader findDocument(Rete engine, String filename)
            throws JessException, IOException {
        PushbackReader fis;
        try {
            if (engine.getApplet() == null)
                fis = new PushbackReader(new FileReader(filename));
            else {
                URL url = new URL(engine.getApplet().getDocumentBase(),
                                  filename);
                fis = new PushbackReader(new InputStreamReader(url.openStream()));
            }
        } catch (Exception e) {
            // Try to find a resource file, too.
            URL u = engine.getResource(filename);
            if (u == null)
                throw new JessException("batch", "Cannot open file", e);

            InputStream is = u.openStream();
            fis = new PushbackReader(new InputStreamReader(is));
        }
        return fis;
    }

    /**
     * Returns true if the first character of the file is a "&lt;".
     * @param fis the open file
     * @return true if the file is XML
     * @throws IOException if anything goes wrong
     */
    public static boolean isXMLDocument(PushbackReader fis) throws IOException {
        int c = ' ';
        while (c != -1 && Character.isWhitespace((char) c))
            c = fis.read();
        if (c != -1)
            fis.unread(c);
        return (c == '<');
    }

    /**
     * The implementation of the Jess language "batch" command, which just calls {@link #batch}.
     * @param vv the function call
     * @param context the execution context
     * @return the last expression parsed
     * @throws JessException if anything goes wrong
     */
    public Value call(ValueVector vv, Context context) throws JessException {
        String filename = vv.get(1).stringValue(context);
        return batch(filename, context.getEngine(), context);
    }
}
