package jess.server;

import jess.*;

import java.io.*;
import java.util.*;

/**
 * This class is part of the implementation of the JessDE
 * debugger. It's responsible for handling breakpoints and other
 * debugger commands.
 * <p/>
 * (C) 2005 Sandia National Laboratories<br>
 */

public class DebugListener extends Observable implements JessListener {
    private boolean m_suspended = false;
    private boolean m_stepping = false;
    private boolean m_firstTime = true;

    public static final String STARTED = "STARTED";
    public static final String STOPPED = "STOPPED";
    public static final String QUIT = "QUIT";
    public static final String SUSPEND = "SUSPENDED";
    public static final String RESUME = "RESUMED";
    public static final String RESUME_STEP = "RESUMED_STEP";
    public static final String SUSPEND_STEP = "SUSPENDED_STEP";
    public static final String SUSPENDED_BREAK = "SUSPENDED_BREAK";
    public static final String STACK = "STACK";
    public static final String ERROR = "ERROR";
    public static final String OK = "OK";
    public static final String PRINT = "PRINT";
    public static final String FACT = "FACT";

    PrintWriter m_eventSink, m_cmdSink;
    BufferedReader m_cmdSource;
    private Map m_breakpoints = new HashMap();
    private Funcall m_stepOverThis;
    private boolean m_listeningForEvents = true;

    public DebugListener(Reader input, Writer output, Writer events) {
        setCmdSource(input);
        setCmdSink(output);
        setEventSink(events);
    }

    public DebugListener() {
    }

    public void eventHappened(JessEvent je) throws JessException {
        if (m_listeningForEvents) {
            if (m_firstTime) {
                m_firstTime = false;
                sendEvent(STARTED);
                setSuspended(true);
            }

            if (processEvent(je)) {
                processCommands(je);
            }
        }
    }

    private boolean processEvent(JessEvent je) throws TerminatedException {
        if (je.getType() == JessEvent.USERFUNCTION_CALLED) {
            handleBreakpoints(je);
            return true;
        } else if (je.getType() == JessEvent.USERFUNCTION_RETURNED) {
            handleSpecialFunctions(je);
            if (je.getObject() == m_stepOverThis) {
                m_stepOverThis = null;
                setStepping(true);
            }
            return false;
        } else
            return false;
    }

    private void handleSpecialFunctions(JessEvent je) throws TerminatedException {
        Funcall funcall = ((Funcall) je.getObject());
        String functionName = funcall.getName();

        if (functionName.equals("exit")) {
            terminate();
        }
    }

    private void terminate() throws TerminatedException {
        sendEvent(QUIT);
        setSuspended(false);
        throw new TerminatedException("DebugListener", "Quit by user command", QUIT);
    }

    private void handleBreakpoints(JessEvent je) {
        if (isSteppingOver())
            return;

        Context context = je.getContext();
        LineNumberRecord lnr = context.getLineNumberRecord();
        if (lnr != null) {
            if (shouldBreakAt(lnr.getFileName(), lnr.getLineno())) {
                sendEvent(SUSPENDED_BREAK + " " + lnr.getLineno() + " " + lnr.getFileName().length());
                m_eventSink.print(lnr.getFileName());
                m_eventSink.flush();

                setSuspended(true);
            }
        }
    }

    private void processCommands(JessEvent je) throws TerminatedException {
        try {
            if (isSteppingOver())
                return;

            if (isSuspended() || isStepping() || commandReady()) {
                do {
                    if (isStepping()) {
                        setStepping(false);
                        sendEvent(SUSPEND_STEP);
                    }

                    String command = readCommand();

                    if (command != null)
                        execute(command, je);
                    else
                        break;

                } while (isSuspended());
            }
        } catch (IOException e) {
            reportError(e);
        }
    }

    private boolean isSteppingOver() {
        return m_stepOverThis != null;
    }

    private synchronized boolean isSuspended() {
        return m_suspended;
    }

    private synchronized boolean isStepping() {
        return m_stepping;
    }

    private synchronized void setSuspended(boolean suspended) {
        m_suspended = suspended;
    }

    private synchronized void setStepping(boolean stepping) {
        m_stepping = stepping;
        m_suspended = !stepping;
    }

    private void execute(String command, JessEvent je) throws TerminatedException {
        try {
            if (command.equals("quit")) {
                terminate();
            } else if (command.equals("suspend") && !isSuspended()) {
                sendEvent(SUSPEND);
                setSuspended(true);
            } else if (command.equals("resume") && isSuspended()) {
                sendEvent(RESUME);
                setSuspended(false);
            } else if (command.equals("step") && isSuspended()) {
                sendEvent(RESUME_STEP);
                setStepping(true);
            } else if (command.equals("stepo") && isSuspended()) {
                m_stepOverThis = (Funcall) je.getObject();
                sendEvent(RESUME_STEP);
                setSuspended(false);
            } else if (command.equals("stack") && isSuspended()) {
                handleStackCommand(je);
            } else if (command.startsWith("print ")) {
                handlePrintCommand(je, command);
            } else if (command.startsWith("break ")) {
                handleBreakCommand(command);
            } else if (command.startsWith("unbreak ")) {
                handleUnbreakCommand(command);
            } else if (command.startsWith("fact ")) {
                handleFactCommand(command, je);
            } else if (command.startsWith("activation")) {
                handleActivationCommand(command, je);
            } else if (command.startsWith("agenda")) {
                handleAgendaCommand(command, je);
            }
            sendResponse(OK);

        } catch (TerminatedException ex) {
            sendResponse(OK);
            throw ex;

        } catch (JessException ex) {
            sendResponse(ERROR + ": " + ex.getMessage());
        }
    }

    private void handleAgendaCommand(String command, JessEvent je) throws JessException {
        StringTokenizer tokens = new StringTokenizer(command, " \t");
        if (tokens.countTokens() > 3)
            throw new JessException("handleAgendaCommand", "Wrong number of arguments", tokens.countTokens());
        Rete engine = je.getContext().getEngine();
        String agendaTokenIgnored = tokens.nextToken();
        String countString = tokens.nextToken();
        String module;
        if (tokens.countTokens() == 3)
            module = tokens.nextToken();
        else
            module = engine.getFocus();
        try {
            int requested = Integer.parseInt(countString);
            Iterator it = engine.listActivations(module);
            StringWriter output = new StringWriter();
            PrintWriter out = new PrintWriter(output, true);
            int count = 0;
            while (count < requested && it.hasNext()) {
                Activation a = (Activation) it.next();
                if (a.isInactive())
                    continue;
                printActivation(a, out);
                ++count;
            }
            m_eventSink.print("AGENDA ");
            m_eventSink.print(module);
            m_eventSink.print(" ");
            m_eventSink.println(count);
            m_eventSink.print(output.toString());
            m_eventSink.flush();

        } catch (NumberFormatException nfe) {
            throw new JessException("handleAgendaCommand", "Bad argument", countString);
        }
    }

    private void handleActivationCommand(String command, JessEvent je) throws JessException {
        StringTokenizer tokens = new StringTokenizer(command, " \t");
        if (tokens.countTokens() != 1)
            throw new JessException("handleActivationCommand", "Wrong number of arguments", tokens.countTokens());
        Rete engine = je.getContext().getEngine();
        Activation activation = engine.getThisActivation();
        if (activation == null) {
            m_eventSink.println("ACTIVATION NONE");
        } else {
            m_eventSink.print("ACTIVATION ");
            printActivation(activation, m_eventSink);
        }
        m_eventSink.flush();
    }

    private void printActivation(Activation activation, PrintWriter out) throws JessException {
        out.print(activation.getRule().getName());
        Token token = activation.getToken();
        final int size = token.size();
        for (int i = 0; i < size; ++i) {
            out.print(" ");
            out.print(new FactIDValue(token.fact(i)));
        }
        out.println();
    }

    private void handleFactCommand(String command, JessEvent je) throws JessException {
        StringTokenizer tokens = new StringTokenizer(command, " \t");
        if (tokens.countTokens() != 2)
            throw new JessException("handleFactCommand", "Wrong number of arguments", tokens.countTokens());
        String factTokenIgnored = tokens.nextToken();
        String factNumberString = tokens.nextToken();
        int factNumber;
        try {
            factNumber = Integer.parseInt(factNumberString);
            Context context = je.getContext();
            Fact fact = context.getEngine().findFactByID(factNumber);
            if (fact == null)
                throw new JessException("handleFactCommand", "No such fact id", factNumber);
            Deftemplate template = fact.getDeftemplate();
            int slotCount = template.getNSlots();
            sendEvent(FACT + " " + factNumber + " " + fact.getName() + " " + slotCount);
            for (int i = 0; i < slotCount; ++i) {
                String slotValue = fact.get(i).toString();
                String slotName = template.getSlotName(i);
                m_eventSink.println(slotName + ' ' + String.valueOf(slotValue.length()));
                m_eventSink.write(slotValue);
                m_eventSink.flush();
            }

        } catch (NumberFormatException nfe) {
            throw new JessException("handleFactCommand", "Bad argument", factNumberString);
        }
    }

    static class BreakCommandData {
        int line;
        String file;

        public BreakCommandData(String aFile, int aLine) {
            file = aFile;
            line = aLine;
        }
    }

    void handleBreakCommand(String command) throws JessException {
        BreakCommandData data = parseBreakCommand(command);
        addBreakpoint(data.file, data.line);
    }

    private BreakCommandData parseBreakCommand(String command) throws JessException {
        String lineNumberToken = null;
        String file;

        try {
            // Strip off command token
            String args = command.substring(command.indexOf(' ') + 1);
            lineNumberToken = args.substring(0, args.indexOf(' '));
            file = args.substring(args.indexOf(' ') + 1);
            int line = Integer.parseInt(lineNumberToken);
            return new BreakCommandData(file, line);
        } catch (NumberFormatException e) {
            throw new JessException("parseBreakCommand", "Invalid line number", lineNumberToken);
        } catch (StringIndexOutOfBoundsException e) {
            throw new JessException("parseBreakCommand", "Wrong number of arguments", command);
        }
    }

    private void handleUnbreakCommand(String command) throws JessException {
        BreakCommandData data = parseBreakCommand(command);
        removeBreakpoint(data.file, data.line);
    }

    static class PrintableData {
        String expr;
        int level;

        PrintableData(int aLevel, String exprValue) {
            level = aLevel;
            expr = exprValue;
        }
    }

    public void handlePrintCommand(JessEvent je, String command) throws JessException {
        PrintableData data = parsePrintCommand(command);
        String result = ERROR;
        int pointer = data.level;
        Context context = findRequestedStackFrame(je, pointer);
        if (context != null) {
            Rete engine = context.getEngine();
            try {
                // TODO Threading issues here
                m_listeningForEvents = false;
                result = engine.eval(data.expr, context).toStringWithParens();
            } finally {
                m_listeningForEvents = true;
            }
        }
        // TODO This isn't going to work -- how to tag these requests?
        sendEvent(PRINT + " " + data.level + " " + data.expr + " " + result.length());
        m_eventSink.print(result);
        m_eventSink.flush();
    }

    private Context findRequestedStackFrame(JessEvent je, int pointer) {
        Context context = je.getContext();
        while (pointer > 0 && context != null) {
            context = context.getParent();
            --pointer;
        }
        return context;
    }

    PrintableData parsePrintCommand(String command) throws JessException {
        StringTokenizer tokens = new StringTokenizer(command, " \t");
        if (tokens.countTokens() != 3)
            throw new JessException("parsePrintData", "Wrong number of arguments", tokens.countTokens());
        String printTokenIgnored = tokens.nextToken();
        int level, count;
        String levelToken = tokens.nextToken();
        try {
            try {
                level = Integer.parseInt(levelToken);
            } catch (NumberFormatException e) {
                throw new JessException("parsePrintData", "Stack level not an integer", levelToken);
            }
            String charCountToken = tokens.nextToken();
            try {
                count = Integer.parseInt(charCountToken);
            } catch (NumberFormatException e) {
                throw new JessException("parsePrintData", "Char count not an integer", charCountToken);
            }
            char[] buffer = new char[count];
            m_cmdSource.read(buffer);
            return new PrintableData(level, new String(buffer));
        } catch (IOException e) {
            throw new JessException("parsePrintCommand", "I/O Error", e);
        }
    }

    private void handleStackCommand(JessEvent je) throws JessException {
        Context context = je.getContext();
        StringBuffer result = new StringBuffer(STACK + " ");

        while (context != null) {
            String fileName = "<unknown>";
            String functionName = "<unknown>";
            int lineNumber = 0;
            Funcall funcall = context.getFuncall();
            LineNumberRecord lineNumberRecord = context.getLineNumberRecord();
            if (lineNumberRecord != null) {
                fileName = lineNumberRecord.getFileName();
                lineNumber = lineNumberRecord.getLineno();
                functionName = funcall.getName();
            }

            result.append(functionName).append(",").append(fileName).append(",").append(lineNumber);
            Set names = new HashSet();
            for (Iterator it = context.getVariableNames(); it.hasNext();) {
                String name = (String) it.next();
                if (!name.startsWith("%"))
                    names.add(name);
            }

            addBindingValues(context, names);
            addDefglobals(context, names);
            ArrayList list = new ArrayList(names);
            Collections.sort(list);
            for (Iterator it = list.iterator(); it.hasNext();) {
                result.append(',');
                result.append(it.next());
            }
            result.append("|");
            context = context.getParent();
        }
        sendEvent(result.toString());
    }

    static void addDefglobals(Context context, Set list) {
        Rete engine = context.getEngine();
        for (Iterator it = engine.listDefglobals(); it.hasNext();) {
            Defglobal global = (Defglobal) it.next();
            list.add(global.getName());
        }
    }

    static void addBindingValues(Context context, Set names) throws JessException {
        Funcall funcall = context.getFuncall();
        if (funcall != null) {
            for (int i = 1; i < funcall.size(); ++i) {
                Value v = funcall.get(i);
                if (v instanceof BindingValue) {
                    BindingValue bindingValue = (BindingValue) v;
                    names.add(bindingValue.getName());
                }
            }
        }
    }

    void setCmdSource(Reader cmdSource) {
        m_cmdSource = new BufferedReader(cmdSource);
    }

    void setCmdSink(Writer cmdSink) {
        m_cmdSink = new PrintWriter(cmdSink, true);
    }

    void setEventSink(Writer eventSink) {
        m_eventSink = new PrintWriter(eventSink, true);
    }

    public void reportError(Throwable e) {
        e.printStackTrace(m_eventSink);
        m_eventSink.flush();
    }

    public void sendEvent(String event) {
        m_eventSink.println(event);
        m_eventSink.flush();
        setChanged();
        notifyObservers(event);
    }

    public void sendResponse(String response) {
        m_cmdSink.println(response);
        m_cmdSink.flush();
    }

    public boolean commandReady() throws IOException {
        return m_cmdSource.ready();
    }

    public String readCommand() throws IOException {
        return m_cmdSource.readLine();
    }

    public boolean waitForConnections(int debugPort, int eventPort) {
        Thread t1 = new CommandSocketConnector(this, debugPort);
        Thread t2 = new EventSocketConnector(this, eventPort);

        try {
            t1.start();
            t2.start();
            t1.join();
            t2.join();

        } catch (InterruptedException e) {
            return false;
        }

        return (m_cmdSink != null && m_cmdSource != null && m_eventSink != null);
    }

    void addBreakpoint(String filename, int lineNumber) {
        File file = new File(filename);
        String tag = file.getName();
        Set lineNumbers = (Set) m_breakpoints.get(tag);
        if (lineNumbers == null) {
            lineNumbers = new HashSet();
            m_breakpoints.put(tag, lineNumbers);
        }
        lineNumbers.add(new Integer(lineNumber));
    }

    public void removeBreakpoint(String filename, int lineNumber) {
        File file = new File(filename);
        String tag = file.getName();
        Set lineNumbers = (Set) m_breakpoints.get(tag);
        if (lineNumbers == null)
            return;
        lineNumbers.remove(new Integer(lineNumber));
    }

    public boolean shouldBreakAt(String filename, int lineNumber) {
        if (filename == null)
            return false;
        File file = new File(filename);
        String tag = file.getName();
        Set lineNumbers = (Set) m_breakpoints.get(tag);
        if (lineNumbers == null)
            return false;
        Integer key = new Integer(lineNumber);
        return lineNumbers.contains(key);
    }
}
