package jess.xml;

import jess.*;

import java.io.*;
import java.util.*;

/**
 * Renders Jess constructs as JessML code. You can use this class in
 * two ways: as a standalone application, it can translate a whole
 * file of Jess code into JessML. Or inside a Java application, you
 * can construct an XMLPrinter with a Jess construct as a constructor
 * argument, then call toString() to get the JessML version. Any
 * construct that implements Visitable can be rendered.
 * <p/>
 * (C) 2005 Sandia National Laboratories<br>
 *
 * @see jess.Visitable
 */
public class XMLPrinter implements Visitor {
    private final Visitable m_visitable;

    // Functions that should be executed during parsing because they affect
    // the outcome of the parse
    private static final Set s_keepers = new HashSet();
    static {
        s_keepers.add("defclass");
        s_keepers.add("do-backward-chaining");
        s_keepers.add("set-current-module");
    }

    /**
     * Create an XMLPrinter to render the given Visitable. When toString() is called on this object,
     * it will return an XML version of the argument.
     * @param v a construct or other Visitable to render
     *
     * @see jess.Visitor
     */
    public XMLPrinter(Visitable v) {
        m_visitable = v;
    }

    public Object visitDeffacts(Deffacts d) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("facts");
        xml.textElementNoSpace("name", d.getName());
        possiblyEmitDocstring(xml, d);
        int count = d.getNFacts();
        for (int i = 0; i < count; ++i) {
            Fact fact = d.getFact(i);
            xml.append(visitFact(fact));
        }
        xml.closeTag("facts");
        return xml.toString();
    }

    private static void possiblyEmitDocstring(XMLWriter xml, Named named) {
        String docstring = named.getDocstring();
        if (docstring != null && docstring.length() > 0) {
            xml.textElementNoSpace("comment", named.getDocstring());
        }
    }

    public Object visitDeftemplate(Deftemplate d) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("template");
        xml.textElementNoSpace("name", d.getName());
        if (d.getParent() != Deftemplate.getRootTemplate()) {
            xml.textElementNoSpace("extends", d.getParent().getName());
        }
        possiblyEmitDocstring(xml, d);

        if (hasTemplateDeclarables(d)) {
            xml.openTag("properties");
            if (d.isShadowTemplate()) {
                emitProperty(xml, "from-class", newSymbolValue(d.getShadowClassName()));
            }
            if (d.isSlotSpecific()) {
                emitProperty(xml, "slot-specific", Funcall.TRUE);
            }
            if (d.getBackwardChaining()) {
                emitProperty(xml, "backchain-reactive", Funcall.TRUE);
            }
            // include-variables?

            xml.closeTag("properties");
        }

        if (!d.isShadowTemplate()) {
            for (int i = 0; i < d.getNSlots(); ++i) {
                try {
                    String name = d.getSlotName(i);
                    Value v = d.getSlotDefault(i);
                    if (d.getSlotType(i) == RU.MULTISLOT) {
                        xml.openTag("multislot");
                        xml.textElementNoSpace("name", name);
                        xml.append(visitList(v.listValue(null)));
                        xml.closeTag("multislot");
                    } else {
                        xml.openTag("slot");
                        int type = d.getSlotDataType(i);
                        if (type != -1) {
                            xml.textElementNoSpace("type", RU.getTypeName(type));
                        }                        xml.textElementNoSpace("name", name);
                        xml.appendln(visitValue(v));
                        xml.closeTag("slot");
                    }

                } catch (JessException e) {
                    return e.toString();
                }
            }
        }
        xml.closeTag("template");
        return xml.toString();
    }

    private static void emitProperty(XMLWriter xml, String name, Value value) {
        xml.openTag("property");
        xml.textElementNoSpace("name", name);
        xml.append(visitValue(value));
        xml.closeTag("property");
    }

    private static Value newSymbolValue(String s) {
        try {
            return new Value(s, RU.SYMBOL);
        } catch (JessException cantHappen) {
            return null;
        }
    }

    private static boolean hasTemplateDeclarables(Deftemplate d) {
        return d.isShadowTemplate() || d.isSlotSpecific() || d.getBackwardChaining();
    }

    public Object visitDeffunction(Deffunction d) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("function");
        xml.textElementNoSpace("name", d.getName());
        possiblyEmitDocstring(xml, d);
        xml.openTag("arguments");
        for (Iterator it = d.getArguments(); it.hasNext();) {
            Deffunction.Argument arg = (Deffunction.Argument) it.next();
            xml.openTag("argument");
            xml.textElementNoSpace("name", arg.getName());
            xml.textElementNoSpace("type", RU.getTypeName(arg.getType()));
            xml.closeTag("argument");
        }
        xml.closeTag("arguments");
        for (Iterator it = d.getActions(); it.hasNext();) {
            xml.appendln(visitValue((Value) it.next()));
        }
        xml.closeTag("function");
        return xml.toString();
    }

    public Object visitDefglobal(Defglobal d) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("global");
        xml.textElementNoSpace("name", d.getName());
        xml.appendln(visitValue(d.getInitializationValue()));
        xml.closeTag("global");
        return xml.toString();
    }

    public Object visitDefrule(Defrule d) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("rule");
        xml.textElementNoSpace("name", d.getName());
        possiblyEmitDocstring(xml, d);
        // TODO Lots more declarables!
        if (hasRuleDeclarables(d)) {
            xml.openTag("properties");
            if (d.hasNonDefaultSalience()) {
                emitProperty(xml, "salience", d.getSalienceValue());
            }
            if (d.getAutoFocus()) {
                emitProperty(xml, "auto-focus", Funcall.TRUE);
            }
            if (d.isNoLoop()) {
                emitProperty(xml, "no-loop", Funcall.TRUE);
            }
            if (d.getNodeIndexHash() != 0) {
                emitProperty(xml, "node-index-hash", makeIntValue(d.getNodeIndexHash()));
            }
            xml.closeTag("properties");
        }
        xml.openTag("lhs");
        if (d.getNext() != null) {
            xml.openTag("group");
            xml.textElementNoSpace("name", "or");
        }
        Defrule rule = d;
        while (rule != null) {
            xml.append(visitGroup((Group) rule.getConditionalElements()));
            rule = (Defrule) rule.getNext();
        }
        if (d.getNext() != null)
            xml.closeTag("group");

        xml.closeTag("lhs");
        xml.openTag("rhs");
        for (int i = 0; i < d.getNActions(); ++i)
            xml.append(visitFuncall(d.getAction(i)));
        xml.closeTag("rhs");
        xml.closeTag("rule");
        return xml.toString();
    }

    private static boolean hasRuleDeclarables(Defrule d) {
        return d.hasNonDefaultSalience() ||
                d.getAutoFocus() ||
                d.isNoLoop() ||
                d.getNodeIndexHash() != 0;
    }

    private static boolean hasQueryDeclarables(Defquery d) {
        return d.getMaxBackgroundRules() > 0 || d.getNodeIndexHash() != 0;
    }

    public Object visitDefquery(Defquery d) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("query");
        xml.textElementNoSpace("name", d.getName());
        possiblyEmitDocstring(xml, d);
        int nvars = d.getNVariables();
        xml.openTag("arguments");
        for (int i = 0; i < nvars; ++i) {
            Variable arg = d.getQueryVariable(i);
            xml.openTag("argument");
            xml.textElementNoSpace("name", arg.variableValue(null));
            xml.textElementNoSpace("type", RU.getTypeName(arg.type()));
            xml.closeTag("argument");
        }
        xml.closeTag("arguments");

        if (hasQueryDeclarables(d)) {
            xml.openTag("properties");
            if (d.getMaxBackgroundRules() > 0) {
                emitProperty(xml, "max-background-rules", makeIntValue(d.getMaxBackgroundRules()));
            }
            if (d.getNodeIndexHash() != 0) {
                emitProperty(xml, "node-index-hash", makeIntValue(d.getNodeIndexHash()));
            }
            xml.closeTag("properties");
        }

        xml.openTag("lhs");
        if (d.getNext() != null) {
            xml.openTag("group");
            xml.textElementNoSpace("name", "or");
        }
        Defquery query = d;
        while (query != null) {
            xml.append(visitGroup((Group) query.getConditionalElements()));
            query = (Defquery) query.getNext();
        }
        if (d.getNext() != null)
            xml.closeTag("group");

        xml.closeTag("lhs");
        xml.closeTag("query");
        return xml.toString();
    }

    private static Value makeIntValue(int i) {
        try {
            return new Value(i, RU.INTEGER);
        } catch (JessException cantHappen) {
            return null;
        }
    }

    public Object visitPattern(Pattern p) {
        if (p.getName().indexOf(Defquery.QUERY_TRIGGER) > 0)
            return "";
        XMLWriter xml = new XMLWriter();
        xml.openTag("pattern");
        String name = p.getName();
        if (name.equals("MAIN::test"))
            name = "test";
        xml.textElementNoSpace("name", name);
        if (p.getBoundName() != null) {
            xml.textElementNoSpace("binding", p.getBoundName());
        }

        Deftemplate dt = p.getDeftemplate();
        try {

            for (int i = 0; i < p.getNSlots(); ++i) {
                if (p.getNTests(i) != 0) {
                    xml.openTag("slot");
                    xml.textElementNoSpace("name", dt.getSlotName(i));
                    for (int j = 0; j < p.getNTests(i); ++j) {
                        Test1 t = p.getTest(i, j);
                        xml.append(visitTest1(t));
                    }
                    xml.closeTag("slot");
                }
            }


        } catch (JessException ex) {
            return ex.toString();
        }

        xml.closeTag("pattern");
        return xml.toString();
    }

    public Object visitGroup(Group p) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("group");
        xml.textElementNoSpace("name", p.getName());

        for (int i = 0; i < p.getGroupSize(); ++i)  {
            ConditionalElement ce = p.getConditionalElement(i);
            if (i ==0 && ce.getName().equals(Deftemplate.getInitialTemplate().getName()))
                continue;
            xml.append(new XMLPrinter((Visitable) ce));
        }
        xml.closeTag("group");
        return xml.toString();
    }

    public Object visitTest1(Test1 t) {
        XMLWriter writer = new XMLWriter();
        writer.openTag("test");
        writer.textElementNoSpace("type", t.getTest() == Test1.EQ ? "eq" : "neq");
        if (t.getConjunction() != RU.NONE) {
            writer.textElementNoSpace("conjunction", t.getConjunction() == RU.AND ? "and" : "or");
        }

        writer.appendln(visitValue(t.getValue()));
        writer.closeTag("test");
        return writer.toString();
    }

    private static Object visitValue(Value value) {
        try {
            if (value.type() == RU.FACT)
                return visitFact(value.factValue(null));
            else if (value.type() == RU.FUNCALL)
                return visitFuncall(value.funcallValue(null));
            else if (value.type() == RU.LIST)
                return makeSlot(value.listValue(null));

            StringBuffer buffer = new StringBuffer();
            buffer.append("<value><type>");
            buffer.append(RU.getTypeName(value.type()));
            buffer.append("</type>");

            switch (value.type()) {
                case RU.SYMBOL:
                case RU.STRING:
                    buffer.append(XMLWriter.escape(value.stringValue(null)));
                    break;
                case RU.VARIABLE:
                case RU.MULTIVARIABLE:
                    buffer.append(value.variableValue(null));
                    break;
                case RU.FLOAT:
                case RU.INTEGER:
                case RU.LONG:
                default:
                    buffer.append(value.toString());
                    break;
            }
            buffer.append("</value>");
            return buffer.toString();

        } catch (JessException e) {
            return e.toString();
        }
    }

    private static Object visitList(ValueVector list) {
        try {
            XMLWriter writer = new XMLWriter();
            for (int i = 0; i < list.size(); ++i)
                writer.appendln(visitValue(list.get(i)));
            return writer.toString();
        } catch (JessException e) {
            return e.toString();
        }
    }

    private static Object makeSlot(ValueVector list) {
        XMLWriter xml = new XMLWriter();
        try {

            xml.openTag("slot");
            if (list.size() > 0) {
                if (list.get(0).type() == RU.SYMBOL)
                    xml.textElementNoSpace("name", list.get(0).symbolValue(null));
                else
                    xml.appendln(visitValue(list.get(0)));
                for (int i = 1; i < list.size(); ++i)
                    xml.appendln(visitValue(list.get(i)));
            }
            xml.closeTag("slot");

            return xml.toString();
        } catch (JessException e) {
            xml.append(e.toString());
            return xml.toString();
        }
    }


    private static Object visitFact(Fact fact) {
        try {
            XMLWriter xml = new XMLWriter();
            Deftemplate dt = fact.getDeftemplate();
            xml.openTag("fact");
            xml.textElementNoSpace("name", fact.getName());
            for (int i = 0; i < dt.getNSlots(); ++i) {
                String name = dt.getSlotName(i);
                Value val = fact.getSlotValue(name);
                if (!dt.getSlotDefault(i).equals(val)) {
                    xml.openTag("slot");
                    xml.textElementNoSpace("name", name);
                    if (val.type() != RU.LIST) {
                        xml.append(visitValue(val));
                        xml.appendln("");
                    } else
                        xml.append(visitList(val.listValue(null)));
                    xml.closeTag("slot");
                }
            }
            xml.closeTag("fact");
            return xml.toString();
        } catch (JessException e) {
            return e.toString();
        }
    }

    private static Object visitFuncall(Funcall funcall) {
        XMLWriter xml = new XMLWriter();
        try {
            xml.openTag("funcall");
            xml.textElementNoSpace("name", funcall.get(0).symbolValue(null));
            for (int i = 1; i < funcall.size(); ++i)
                xml.appendln(visitValue(funcall.get(i)));
            xml.closeTag("funcall");
            return xml.toString();
        } catch (JessException e) {
            return e.toString();
        }
    }

    public Object visitAccumulate(Accumulate accumulate) {
        return notImplemented(accumulate);
    }

    public Object visitDefmodule(Defmodule defmodule) {
        XMLWriter xml = new XMLWriter();
        xml.openTag("module");
        xml.textElementNoSpace("name", defmodule.getName());
        possiblyEmitDocstring(xml, defmodule);
        xml.closeTag("module");
        return xml.toString();
    }

    private static String notImplemented(Visitable visitable) {
        return "<!--\n  NOT IMPLEMENTED\n" +
                new PrettyPrinter(visitable).toString() +
                "\n-->";
    }

    public String toString() {
        return (String) m_visitable.accept(this);
    }

    /**
     * Reads a file of Jess code and emits equivalent XML on standard output.
     * @param args first argument is the path to a file of Jess code
     * @throws FileNotFoundException if the file doesn't exist
     * @throws JessException if anything goes wrong
     */
    public static void main(String[] args) throws FileNotFoundException, JessException {
        FileReader fr = new FileReader(args[0]);
        Rete r = new Rete();
        StringWriter devNull = new StringWriter();
        r.addOutputRouter("t", devNull);
        r.addOutputRouter("WSTDOUT", devNull);
        r.addOutputRouter("WSTDERR", devNull);

        Jesp j = new Jesp(fr, r);
        j.setFileName(args[0]);
        Object o;

        System.out.println("<?xml version='1.0' encoding='US-ASCII'?>");
        System.out.println("<?JessML-version 1?>");
        System.out.println("<rulebase>");


        do {
            o = j.parseExpression(r.getGlobalContext(), false);
            if (o instanceof Visitable) {
                XMLPrinter p = new XMLPrinter((Visitable) o);
                System.out.println(p.toString());

            } else if (o instanceof List) {
                List v = (List) o;
                for (int i = 0; i < v.size(); i++) {
                    XMLPrinter p = new XMLPrinter((Visitable) v.get(i));
                    System.out.println(p.toString());
                }

            } else if (o instanceof FuncallValue) {
                Funcall funcall = ((FuncallValue) o).funcallValue(null);
                if (shouldExecute(funcall))
                    funcall.execute(r.getGlobalContext());
                System.out.println(visitFuncall(funcall));
            }
        } while (!Funcall.EOF.equals(o));

        System.out.println("</rulebase>");
    }

    private static boolean shouldExecute(Funcall funcall) {
        return s_keepers.contains(funcall.getName());
    }
}
